from __future__ import annotations

import logging
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Dict, Iterable, Optional

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from .. import options_parser
from . import extension
from .querycontainer import IqlQueryContainer

logger = logging.getLogger(__name__)


def get_subqueries_flat(queries: Iterable[SubQuery]) -> Iterable[SubQuery]:
    subqueries_flat = [q_inner for q_outer in queries for q_inner in q_outer.get_subqueries_flat()]
    return subqueries_flat


@dataclass
class SubQuery:
    extension: extension.IqlExtension
    subquery: str
    name: str
    dataframe: Optional[pd.DataFrame] = field(default=None, init=False)
    options: Dict[str, object] = field(default_factory=dict, init=False)
    input_data: object = field(default=None, init=False)
    iqc: IqlQueryContainer

    local_dfs: Dict[str, object] = field(default_factory=dict, init=False)

    def get_subqueries_flat(self) -> Iterable[SubQuery]:
        return [self]

    def merge(self):
        # Nothing to do for a single subquery
        pass

    def get_cache_key(self) -> str:
        """Excludes paramquery and paramquerybatch and paramlist"""
        keypart1 = f"{self.extension.keyword}.{self.extension.subword}"

        if self.options is not None and len(self.options) > 0:
            options_clean = {
                k: v
                for k, v in self.options.items()
                if k != "paramquery" and k != "paramquerybatch" and k != "paramlist"
            }

            keypart2 = str(
                options_clean
            )  # to include any dynamic values, populated from paramquery and paramquerybatch
        else:
            keypart2 = ""

        cache_key = keypart1 + keypart2
        return cache_key

    def get_query(self) -> Optional[str]:
        """Returns the first parameter to:
        keyword(query, *params)"""
        if len(self.options) == 0:
            raise ValueError(f"Options not properly passed or parsed ({self.subquery})")

        value, query = next(iter(self.options.items()))

        if query is None:  # unnamed parameter
            return value
        else:
            return query  # type: ignore

    def __post_init__(self):
        try:
            if self.subquery is not None:
                self.options: Dict[str, object] = options_parser.options_to_list(self.subquery)
        except Exception as e:
            logger.exception("Exception in %.50s", self.subquery)
            raise ValueError(f"Parse Exception of {str(self.subquery)[:50]}") from e


@dataclass
class SubQueryGroup(SubQuery):
    """SubQueryGroups represent a set of queries whose outputs will be combined.
    SubQueries are executed in parallel (via get_subqueries_flat),
    so this step is needed to gather the results.

    Our first implementation was to rely on database Unions, but performance wasn't
    great.

    Future: SubQueryGroups could support nesting, only one level supported now."""

    subqueries: Iterable[SubQuery]

    def __post_init__(self):
        self.subquery = None  # type: ignore
        self.options = None  # type: ignore

    def get_subqueries_flat(self) -> Iterable[SubQuery]:
        return self.subqueries

    def get_query(self) -> None:
        raise NotImplementedError("SubQueryGroups do not have/need a query")

    def merge(self):
        """Doesn't handle mismatched schemas when merging parquet files."""
        if self.extension.use_path_replacement():
            files = [  # type: ignore
                self.extension.get_path_replacement(sq, quote=False) for sq in self.subqueries
            ]

            logger.debug("Merging %s parquet files: %s", len(files), files)

            tables = [pq.read_table(f) for f in files]
            logger.debug("Done reading, concatting")

            table = harmonize_and_concat_arrow_tables(tables)

            outfile = self.extension.get_path_replacement(self, quote=False)
            logger.debug("Writing to final parquet file: %s", outfile)

            pq.write_table(table, outfile)

        else:
            dfs: list[pd.DataFrame] = [sq.dataframe for sq in self.subqueries]  # type: ignore
            logger.debug("Concatting %s", len(dfs))
            if len(dfs) == 1:
                self.dataframe = dfs[0]
            else:
                arrow_tables = [df for df in dfs if isinstance(df, pa.Table)]

                if len(arrow_tables) == len(dfs):
                    logger.debug("Merging py arrow tables")
                    for df in dfs:
                        if df is not None:
                            logger.debug("Table len: %s", df.num_rows)
                    df_combined = harmonize_and_concat_arrow_tables(dfs)
                    logger.debug("Merged len: %s", len(df_combined))
                    self.dataframe = df_combined
                elif len(arrow_tables) > 0:
                    raise ValueError("Mix of arrow and pandas dfs, not supported")
                else:
                    self.dataframe = pd.concat(dfs)

        logger.debug("Done merging subqueries")


def harmonize_and_concat_arrow_tables(tables: Iterable):
    """Harmonize means: if the tables have columns with different dtypes, cast them to a String"""

    # Could also do a Count here, just counting the occurrences of columns, and if the dtypes differ, use a String
    columns = defaultdict(set)
    i = 0

    for t in tables:
        i += 1
        for c in t.schema.names:
            columns[c].add(t.column(c).type)

    if i == 0:
        raise ValueError("No tables passed")

    cast_columns = []
    for column, types in columns.items():
        # if there are multiple column types, ignore the null type
        if len(types) > 1 and len([t for t in types if not pa.types.is_null(t)]) > 1:
            cast_columns.append(column)

    new_tables = []
    for t in tables:
        for c in cast_columns:
            if c in t.schema.names:
                logger.debug("Harmonizing arrow table: Casting %s to string", c)

                idx = t.schema.get_field_index(c)
                t = t.set_column(idx, c, t.column(c).cast("string"))
        new_tables.append(t)

    combined_table = pa.concat_tables(new_tables, promote_options="default")
    return combined_table



import logging
import threading
import time
from typing import Optional

import bql  # type: ignore - not available unless using simulator (dev) or in Bqnt

logger = logging.getLogger(__name__)

_BQSERVICE: Optional[bql.Service] = None

_BPSESSION = None
BQL_LOCK = threading.Lock()  # Prevent concurrent

BQSERVICE_TIMEOUT = 120
BQSERVICE_NEEDED_RETRY = False


def get_bqapi_session():
    global _BPSESSION
    if _BPSESSION is not None:
        return _BPSESSION
    with BQL_LOCK:
        # make sure we didn't create it after acquiring lock
        if _BPSESSION is not None:
            return _BPSESSION
        import bqapi as bp  # pyright: ignore

        _BPSESSION = bp.get_session_singleton()
        return _BPSESSION


class BqServiceRetriever(threading.Thread):
    service = None

    def run(self):
        self.service = bql.Service()


def get_bqservice(retries: int = 3) -> bql.Service:
    global _BQSERVICE
    global BQSERVICE_NEEDED_RETRY

    try:
        if _BQSERVICE is None:  # type: ignore
            logger.debug("Creating new bqService")

            start_time = time.time()

            retriever_thread = BqServiceRetriever()
            retriever_thread.start()
            retriever_thread.join(timeout=BQSERVICE_TIMEOUT)
            if retriever_thread.is_alive():
                raise TimeoutError(f"Unable to get bqservice in {BQSERVICE_TIMEOUT}")
            _BQSERVICE = retriever_thread.service

            end_time = time.time()
            duration = end_time - start_time
            logger.debug("Creating new bqService took %s seconds", round(duration))

    except Exception as e:
        logger.exception("Unable to create bq service")
        if retries <= 0:
            raise e

    if _BQSERVICE is None:
        if retries > 0:
            BQSERVICE_NEEDED_RETRY = True
            logger.warning("Retry get bqservice (retries left %s)", retries)
            return get_bqservice(retries - 1)
        raise ValueError("Unable to obtain bql.Service()")
    return _BQSERVICE


def close_bqservice():
    """Closing the session is not normally needed, but there is a limit on the number of concurrent connections.
    BQL uses (as far as we can tell) a Singleton so multiple bql sessions requests will not consume more connections.
    """
    try:
        global _BQSERVICE
        if _BQSERVICE is None or _BQSERVICE._Service__bqapi_session is None:  # type: ignore
            logger.debug("_Service__bqapi_session already closed or None")
            return
        else:
            logger.debug("Closing _Service__bqapi_session session")
            _BQSERVICE._Service__bqapi_session.close()  # type: ignore
            _BQSERVICE = None
    except Exception:
        logger.exception("Error closing bqapi session")



import asyncio
import hashlib
import logging
import time
from dataclasses import dataclass, field
from pathlib import Path

import cachetools
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq

from . import extension
from .subquery import SubQuery

DEFAULT_CACHE_DIR = "."  # c:\\git\\iql_cache


def set_cache_dir(dir: str):
    global DEFAULT_CACHE_DIR

    DEFAULT_CACHE_DIR = dir


def get_cache_dir():
    return Path(DEFAULT_CACHE_DIR)


logger = logging.getLogger(__name__)


def iql_cache(func):
    if asyncio.iscoroutinefunction(func):

        async def wrapper_a(ext: "extension.IqlExtension", sq: SubQuery, *args, **kwargs):
            logger.debug("Wrapper_a running %s", type(ext.cache))
            cached_result = ext.cache.get(sq) if ext.cache else None
            if cached_result is not None:
                return cached_result
            else:
                start = time.perf_counter()
                new_result = await func(ext, sq, *args, **kwargs)
                end = time.perf_counter()
                if ext.cache:
                    ext.cache.save(sq=sq, data=new_result, cost=end - start)
                return new_result

        return wrapper_a
    else:

        def wrapper(ext: "extension.IqlExtension", sq: SubQuery, *args, **kwargs):
            logger.debug("Wrapper running %s", type(ext))

            cached_result = ext.cache.get(sq) if ext.cache else None
            if cached_result is not None:
                return cached_result
            else:
                start = time.perf_counter()
                new_result = func(ext, sq, *args, **kwargs)
                end = time.perf_counter()
                if ext.cache:
                    ext.cache.save(sq=sq, data=new_result, cost=end - start)
                return new_result

        return wrapper


class SqCache:
    def save(self, sq: SubQuery | str, data: object, cost: float | None):
        """Save the element to cache.

        The cache implementation *may* use cost to decide on whether to cache an item.

        Args:
            sq (SubQuery | str): Subquery (or a cache key, if you want to use an explicit key)
            data (object): Data to be cached.
            cost (float | None): Usually the number of seconds the data took to generate
        """
        ...

    def get(self, sq: SubQuery | str) -> object: ...

    def clear(self, sq: SubQuery | str): ...

    def clear_all(self): ...


@dataclass
class NoopSqCache(SqCache):
    """Caches Nothing"""

    def save(self, sq: SubQuery | str, data: object, cost: float | None):
        pass

    def get(self, sq: SubQuery | str) -> object:
        return None

    def clear(self, sq: SubQuery | str):
        pass

    def clear_all(self):
        pass


@dataclass
class MemoryCache(SqCache):
    """Simple in memory cache. If max_age is None, then caches forever. Otherwise, uses a TTN cache"""

    max_age: int | None = field(default=None, init=True)
    min_cost: int = field(default=-1, init=True)
    _cache: dict = field(default_factory=dict, init=False)

    def __post_init__(self):
        if self.max_age is None:
            self._cache = {}  # just use a dictionary
        else:
            self._cache = cachetools.TTLCache(maxsize=float("inf"), ttl=self.max_age)

    def save(self, sq: SubQuery | str, data: object, cost: float | None):
        if cost is None or cost > self.min_cost:
            key = sq if isinstance(sq, str) else sq.get_cache_key()
            self._cache[key] = data

    def get(self, sq: SubQuery | str) -> object:
        key = sq if isinstance(sq, str) else sq.get_cache_key()
        return self._cache.get(key, None)

    def clear(self, sq: SubQuery | str):
        key = sq if isinstance(sq, str) else sq.get_cache_key()
        del self._cache[key]

    def clear_all(self):
        self._cache = {}


@dataclass
class MemoryAndFileCache(SqCache):
    """Simple in memory cache that ignores age: if age is None, then doesn't cache, otherwise caches forever"""

    max_age: int
    min_cost: int = field(default=-1, init=True)
    cache_dir: Path = field(default_factory=get_cache_dir, init=True)
    return_pyarrow_table: bool = False

    _mem_cache: cachetools.TTLCache = field(init=False)

    def __post_init__(self):
        if not self.cache_dir.exists():
            raise ValueError(f"{self.cache_dir=} does not exist")

        self.clear_all()

    def to_filename(self, key: str):
        filename_hash = hashlib.sha256(key.encode()).hexdigest()
        return self.cache_dir / f"{filename_hash}.parquet"

    def save(self, sq: SubQuery | str, data: object, cost: float | None):
        if cost is None or cost > self.min_cost:
            key = sq if isinstance(sq, str) else sq.get_cache_key()
            logger.debug("Saving %s, cost=%s", key, cost)

            self._mem_cache[key] = data

            outfile = self.to_filename(key)
            if isinstance(data, pd.DataFrame):
                data.to_parquet(outfile)
            elif isinstance(data, pa.Table):
                pq.write_table(data, outfile)
            else:
                raise ValueError(f"Unsupported cache data type {type(data)}")

    def get(self, sq: SubQuery | str) -> object:
        key = sq if isinstance(sq, str) else sq.get_cache_key()

        data = self._mem_cache.get(key, None)
        if data is not None:
            logger.debug("Mem cache hit: %s", key)

            return data
        else:
            outfile = self.to_filename(key)
            if outfile.exists():
                logger.debug("File cache hit: %s", key)
                if (time.time() - outfile.stat().st_mtime) < self.max_age:
                    if self.return_pyarrow_table:
                        return pq.read_table(outfile)
                    else:
                        return pd.read_parquet(outfile)
                else:
                    logger.debug("File cache is expired %s", outfile)
                    return None
            else:
                return None

    def clear(self, sq: SubQuery | str):
        key = sq if isinstance(sq, str) else sq.get_cache_key()
        del self._cache[key]

    def clear_all(self):
        self._mem_cache = cachetools.TTLCache(maxsize=float("inf"), ttl=self.max_age)


@dataclass
class QueryInvalidationCache(SqCache):
    """Caches until the query returns a new value"""

    cache_query: str
    _cache_key: object | None = None
    _cache: dict = field(default_factory=dict, init=False)

    def __post_init__(self):
        self._cache = {}  # just use a dictionary

    def check_cache_valid(self) -> bool:
        from .. import ql

        status = ql.executedf(self.cache_query)
        if not (len(status) and len(status.columns) == 1):
            raise ValueError("Invalid cache")

        if status.iloc[0, 0] == self._cache_key:
            logger.debug("Cache value hasn't changed, cache is valid")
            return True
        else:
            self.clear_all()
            self._cache_key = status.iloc[0, 0]
            logger.debug("Cache value has changed, cache is invalid")
            return False

    def save(self, sq: SubQuery | str, data: object, cost: float | None):
        key = sq if isinstance(sq, str) else sq.get_cache_key()
        self._cache[key] = data

    def get(self, sq: SubQuery | str) -> object:
        if self.check_cache_valid():
            key = sq if isinstance(sq, str) else sq.get_cache_key()
            return self._cache.get(key, None)
        else:
            return None

    def clear(self, sq: SubQuery | str):
        key = sq if isinstance(sq, str) else sq.get_cache_key()
        del self._cache[key]

    def clear_all(self):
        self._cache = {}
        self._cache_key = None



# Autocompletion using the Autocompletion V2: introduced in Ipython 8.6.0


import logging
import re

import sqlparse
from IPython.core.completer import (
    Completion,
    CompletionContext,
    IPCompleter,
    SimpleCompletion,
    SimpleMatcherResult,
    context_matcher,
)

from .. import ql
from .autocomplete_settings import (
    get_column_names,
    get_function_names,
    get_table_names,
    pragma_phrases,
    sql_expects_tablename,
    sql_phrases,
)

logger = logging.getLogger(__name__)

pragma_before = re.compile(r"(?si).*pragma\s*")


def get_aliases(sql: str) -> dict:
    stmt = sqlparse.parse(sql)[0]

    aliases = {
        token.get_alias(): token.get_real_name() for token in stmt.tokens if isinstance(token, sqlparse.sql.Identifier)
    }
    return aliases


class IqlCustomCompleter(IPCompleter):
    def __init__(self, *args, **kwargs):
        self.ipython = kwargs.get("shell")
        super().__init__(*args, **kwargs)

    all_phrases = sql_expects_tablename + sql_phrases + pragma_phrases
    lastword_pat = re.compile(r"(?si)(^|.*[\s])(\S+)\.")
    expects_table_pat = re.compile(r"(?si).*from")

    def convert_to_return(self, cursor_pos, completions: list[str], matched_fragment: str) -> SimpleMatcherResult:
        simple_completions = [
            SimpleCompletion(text=t, type="duckdb") for t in completions if t.startswith(matched_fragment)
        ]

        # It took way too many hours to figure out that matched_fragment="" was needed here
        # Otherwise the results get suppressed

        r = SimpleMatcherResult(
            completions=simple_completions,
            suppress=True,
            matched_fragment=matched_fragment,
            ordered=True,
        )

        return r

    def line_completer_inner(self, event: CompletionContext) -> SimpleMatcherResult:
        # if not %dql, returns nothing
        # if ends with a sql_expects_tablename phrase, then returns just table names
        # if ends in a period, checks to see if prefix is a tablename, and if so, returns column names
        # otherwise returns all sql phrases and tablenames
        # https://github.com/ipython/ipython/blob/a418f38c4f96de1755701041fe5d8deffbf906db/IPython/core/completer.py#L563

        cursor_pos = event.cursor_position
        try:
            logger.debug("%s, %s, %s", type(event), self.ipython, event)

            if hasattr(event, "full_text"):
                text = event.full_text
            else:
                logger.debug("No full_text, nothing to do %s", event)
                return self.convert_to_return(cursor_pos=cursor_pos, completions=[], matched_fragment="")

            if not text.startswith("%iql") and not text.startswith("%%iql"):
                return self.convert_to_return(cursor_pos=cursor_pos, completions=[], matched_fragment="")

            if hasattr(event, "token"):
                token = event.token
            else:
                token = ""

            prev_char_pos = event.cursor_position - len(token) - 1
            prev_char = event.line_with_cursor[prev_char_pos] if prev_char_pos < len(event.line_with_cursor) else None
            line_after = text[4:].strip()

            logger.debug("%s: %s", event, type(event))

            if token is not None and token.endswith("."):
                # VScode is ignoring or suppressing completions after a period.
                # get the word preceding the period
                tablename = token[:-1]

                # TODO: Deal with Aliases and SubQueries (xyz as abc)
                tablename = get_aliases(event.full_text).get(tablename, tablename)
                logger.debug(tablename)

                columns = get_column_names(self.ipython, tablename)
                logger.debug("Using columns: %s", columns)
                return self.convert_to_return(cursor_pos=cursor_pos, completions=columns, matched_fragment="")
            elif token is not None and prev_char == "$":
                logger.debug("Handling substitutions")
                substitutions = list(ql._SUBSTITUTIONS.keys())

                return self.convert_to_return(
                    cursor_pos=cursor_pos, completions=substitutions, matched_fragment=event.token
                )

            # if the last phrase should be followed by a table name, return the list of tables
            elif self.expects_table_pat.match(line_after) is not None:
                names = get_table_names(self.ipython)
                if len(names) == 0:
                    names = ["No Tables or DataFrames Found"]
                logger.debug("Expects table name, returning: %s", names)
                return self.convert_to_return(cursor_pos=cursor_pos, completions=names, matched_fragment=event.token)
            else:
                # default: return all phrases and tablenames
                allp = self.all_phrases + get_table_names(self.ipython) + get_function_names(self.ipython, event.token)
                return self.convert_to_return(cursor_pos=cursor_pos, completions=allp, matched_fragment=event.token)

        except Exception:
            logger.exception("Error completing %s", event)
            return self.convert_to_return(cursor_pos=cursor_pos, completions=[], matched_fragment="")

    @context_matcher()  # type: ignore
    def line_completer(self, event: CompletionContext) -> list[Completion]:
        result = self.line_completer_inner(event=event)
        logger.debug("Result from completion %s", result)
        return result


def init_completer(ip):
    iql_completer = IqlCustomCompleter(shell=ip, greedy=True)
    # Development notes:
    # This took a while to figure out, partially because of the multiple APIs and signatures involved.
    #
    # What I learned:
    # The ipython (get_ipython) environment has a Completer object
    # Completer is an IPCompleter
    # You can extend this completer in several ways, including:
    # - Wrapping it
    # - Registering a custom_completer, of which there are several ways: add_hook, Completer.custom_completers.add_s and add_re, or customer_completers.insert
    #
    # ipython 8.6.0 introduced new a v2 version of the API, for completers decorated with @context_matcher(), this API uses a different signature
    # see ipython.core.completer for more info
    # The main advantage of v2 is getting the full cell text, not just current cell
    #
    # add_s or add_re calls with two parameters:
    # callable(self, event)
    # The event contains: 2023-04-02 07:14:42,857 - magic_duckdb - INFO - namespace(line='%dql asda', symbol='asda', command='%dql', text_until_cursor='%dql asda')
    #
    # set_custom_completer was an alternative method of adding, which seemed inconsistent within VScode... but it passed three parameters:
    # self, ipcompleter, linebuffer, cursor_pos
    #
    # the third method was add_hook('custom_complete'...)
    # which is a synonym for add_s or add_re
    #
    # https://github.com/ipython/ipython/pull/13745
    # https://ipython.readthedocs.io/en/stable/api/generated/IPython.core.completer.html
    # https://raw.githubusercontent.com/ipython/ipython/main/IPython/core/completer.py
    # https://ipython.org/ipython-doc/rel-0.12.1/api/generated/IPython.utils.strdispatch.html#IPython.utils.strdispatch.StrDispatch.s_matches
    # https://github.com/ipython/ipython/blob/9663a9adc4c87490f1dc59c8b6f32cdfd0c5094a/IPython/core/tests/test_completer.py
    #
    # Also, annoyingly, VScode inserts some completions in: https://stackoverflow.com/questions/72824819/vscode-autocomplete-intellisense-in-jupyter-notebook-when-starting-string-lit

    ip.Completer.custom_completer_matcher = iql_completer.line_completer



from ._version import __version__  # noqa: I001
from .ql import (
    configure,
    executedf,
    execute,
    execute_debug,
    get_extension,
    list_extensions,
    register_extension,
    register_alias,
    patch_sqlparse,
)

# Needed for extensions
from .datamodel.extension import IqlExtension
from .datamodel.subquery import SubQuery

from .datamodel.cache import SqCache, set_cache_dir, iql_cache

from .jupyter_magics.iql_magic_ext import load_ipython_extension


__all__ = [
    "__version__",
    "configure",
    "executedf",
    "execute",
    "execute_debug",
    "get_extension",
    "list_extensions",
    "register_extension",
    "load_ipython_extension",
    "register_alias",
    "IqlExtension",
    "SubQuery",
    "SqCache",
    "set_cache_dir",
    "iql_cache",
]

patch_sqlparse()



import logging
from dataclasses import dataclass
from typing import Iterable

import pandas as pd

from ..datamodel.database import IqlDatabase
from ..datamodel.extension import IqlExtension
from ..datamodel.result import IqlResult, IqlResultData
from ..datamodel.subquery import SubQuery
from ..ql import register_extension

logger = logging.getLogger(__name__)


def get_data(sq: SubQuery, source_query) -> pd.DataFrame:
    db: IqlDatabase = sq.iqc.db  # type: ignore
    local_results: Iterable[IqlResult] = [IqlResultData(_data=ldf, name=k, query=k) for k, ldf in sq.local_dfs]

    result = db.execute_query(  # type: ignore
        query=source_query, completed_results=local_results, context=sq.iqc.context
    )
    if result is None:
        raise ValueError(f"Could not get data for query {source_query}")
    else:
        return result.df()  # type: ignore


def get_query_from_options(options) -> str:
    sql = options.get("sql")
    if sql is not None:
        return sql

    dfname = options.get("df")
    tablename = options.get("table")

    if dfname is not None:
        return f"select * from {dfname}"
    elif tablename is not None:
        return f"select * from {tablename}"

    dataname = options.get("data")
    if dataname is not None:
        if " " in dataname:
            return dataname
        else:
            return f"select * from {dataname}"

    raise ValueError("df, table, sql and data are not specified")


@dataclass
class PandasExtension(IqlExtension):
    keyword: str

    def executeimpl(self, sq: SubQuery) -> pd.DataFrame:
        query = get_query_from_options(sq.options)

        return get_data(sq, query)


@dataclass
class PandasTransposeExtension(IqlExtension):
    keyword: str

    def executeimpl(self, sq: SubQuery) -> pd.DataFrame:
        query = get_query_from_options(sq.options)

        result_df = get_data(sq, query)
        cols = sq.options["columns"]
        rows = sq.options["rows"]

        melt_cols = []

        if isinstance(cols, list):
            melt_cols += cols
        elif cols is not None:
            melt_cols.append(cols)

        if isinstance(rows, list):
            melt_cols += rows
        elif isinstance(rows, str):  # str
            melt_cols.append(rows)
            rows = [rows]
        else:
            raise ValueError(f"Unexpected type for rows: {type(rows)}")

        rows.append("name")

        logger.debug(
            "Melting table and repivoting: cols=%s, rows=%s, melt_cols=%s",
            cols,
            rows,
            melt_cols,
        )
        melted_df = (
            result_df.melt(melt_cols, var_name="name")  # noqa
            .pivot(index=rows, columns=cols, values="value")
            .reset_index()
        )
        return melted_df


@dataclass
class PandasReadExtension(IqlExtension):
    keyword: str

    def executeimpl(self, sq: SubQuery) -> pd.DataFrame:
        file = sq.options["file"]
        type = sq.options.get("type", None)

        sheet: str | None = sq.options.get("sheet", None)  # type: ignore

        logger.info("Reading %s from %s", file, type)

        if type == "xls" or type == "xlsx" or type == "excel" or file.endswith(".xls") or file.endswith(".xlsx"):
            if sheet is not None:
                r = pd.read_excel(file, sheet_name=sheet, dtype=str)

                if isinstance(r, dict):
                    to_concat = []
                    for sheet_name, sheet_df in r.items():
                        sheet_df["sheet_name"] = sheet_name

                        to_concat.append(sheet_df)

                    df = pd.concat(to_concat)
                else:
                    df = r

            else:
                df = pd.read_excel(file)

            df["filename"] = file
            return df
        else:
            raise ValueError(f"Unexpected type {type}")


@dataclass
class PandasCrossCorrExtension(IqlExtension):
    keyword: str

    def executeimpl(self, sq: SubQuery) -> pd.DataFrame:
        query = get_query_from_options(sq.options)

        df = get_data(sq, query)

        index = sq.options["index"]
        columns = sq.options["columns"]
        values = sq.options["values"]
        wide = sq.options.get("wide", True)
        min_matches = sq.options.get("min_matches", 5)

        df_p = df.pivot_table(index=index, columns=columns, values=values)  # type: ignore

        if min_matches:
            mask = df_p.notna().astype(int).T.dot(df_p.notna().astype(int)) >= min_matches
            cdf = df_p.corr().where(mask)
        else:
            cdf = df_p.corr()

        if wide:
            return cdf.reset_index()  # type: ignore
        else:
            cmdf = cdf.melt(var_name="var", ignore_index=False).reset_index()
            return cmdf.reset_index()  # type: ignore


def register(keyword: str):
    extension = PandasExtension(keyword=keyword)
    register_extension(extension)

    extension = PandasExtension(keyword=keyword, subword="pivot")
    register_extension(extension)

    extension = PandasTransposeExtension(keyword=keyword, subword="transpose")
    register_extension(extension)

    extension = PandasReadExtension(keyword=keyword, subword="read")
    register_extension(extension)

    extension = PandasCrossCorrExtension(keyword=keyword, subword="crosscorr")
    register_extension(extension)



import logging
from typing import Optional

logger = logging.getLogger(__name__)

# Reference: https://data.bloomberglp.com/professional/sites/10/2017/03/BLPAPI-Core-Developer-Guide.pdf

# API Docs (check for latest): https://bloomberg.github.io/blpapi-docs/


def get_blp_api_session(svc_type):
    # defer blpapi load until this is actually used to prevent loading on init
    import blpapi

    session = blpapi.Session()  # type: ignore

    session.start()
    session.openService(svc_type)
    # TODO Validate the session was successful, since in Bquant Enterprise, this will fail.

    return session


def await_response(session):
    import blpapi

    response = []
    msgidx = 0
    while True:
        ev = session.nextEvent(500)
        # for msg in ev:
        #    logging.debug(f"{msgidx=}: {ev.eventType()=}, {msg=}")

        msgidx = msgidx + 1

        if ev.eventType() == blpapi.Event.RESPONSE or ev.eventType() == blpapi.Event.PARTIAL_RESPONSE:  # type: ignore
            for msg in ev:
                response.append(msg)  # pyright: ignore

            if ev.eventType() == blpapi.Event.RESPONSE:  # type: ignore
                break
    return response


def get_mktdata(field_name, equity="IBM US Equity"):
    """Used when subscribing to streaming real-time and delayed market data."""

    raise NotImplementedError("Not Implemented")


def get_user_entitlements(uuid):
    # https://data.bloomberglp.com/professional/sites/4/blpapi-developers-guide-2.54.pdf
    svc_type: str = "//blp/apiauth"
    session = get_blp_api_session(svc_type)
    ref_svc = session.getService(svc_type)
    request = ref_svc.createRequest("UserEntitlementsRequest")

    userinfo = request.getElement("userInfo")  # type: ignore
    userinfo.setElement("uuid", uuid)  # type: ignore

    session.sendRequest(request)

    response = await_response(session)

    # Convert response to a Dict
    results = []
    for e in response:
        e_py = e.toPy()
        results.append(e_py)

    session.stop()
    return results


#  Request request = authSvc.CreateRequest("UserEntitlementsRequest");
# Element userinfo = request.GetElement("userInfo");
# userinfo.SetElement("uuid", 11223344);


param_map = {
    "per": "periodicitySelection",
    "days": "nonTradingDayFillOption",
    "fill": "nonTradingDayFillMethod",
}


def get_refdata(  # noqa: C901
    fields: tuple[str],
    securities: tuple[str] = ("IBM US Equity",),
    historical: bool = False,
    parameters: Optional[dict] = None,  # Really a frozendict
):
    """_summary_
        Typed as tuples for caching.
        Do not call directly, choose the cached or uncached versions
        Example:
            ba.get_refdata(["PX_LAST"], historical=True, parameters = {"startDate": "20230101", "endDate": "20230103"})
    val

        Args:
            fields (list[str]): _description_
            equities (list[str], optional): _description_. Defaults to ['IBM US Equity'].
            historical (bool, optional): _description_. Defaults to False.
            parameters (dict, optional): _description_. Defaults to {}.

        Returns:
            _type_: _description_
    """

    svc_type: str = "//blp/refdata"
    session = get_blp_api_session(svc_type)
    ref_svc = session.getService(svc_type)

    if historical:
        request = ref_svc.createRequest("HistoricalDataRequest")
    else:
        request = ref_svc.createRequest("ReferenceDataRequest")

    for f in fields:
        request.getElement("fields").appendValue(f)  # type: ignore

    for eq in securities:
        request.getElement("securities").appendValue(eq)  # type: ignore

    if parameters:
        for k, v in parameters.items():
            if k in param_map:
                k = param_map[k]
                v = v.upper()

            request.set(k, v)  # type: ignore

    session.sendRequest(request)

    response = await_response(session)

    # Convert response to a Dict
    results = []
    for e in response:
        e_py = e.toPy()
        logger.debug(e_py)
        if "responseError" in e_py:
            # Handle error
            result = {"status": "responseError"}
            result.update(e_py["responseError"])
            results.append(result)
            continue

        # for security_data in e_py["securityData"]:
        security_data = e_py["securityData"]

        if not isinstance(
            security_data, list
        ):  # historical data is a single security data, with a list of field datas. Ref data is a list of security data and a single field data
            security_data = [security_data]

        for security_data_instance in security_data:
            field_data = security_data_instance["fieldData"]

            if not isinstance(field_data, list):
                field_data = [field_data]

            for field_data_instance in field_data:
                result = {
                    "security": security_data_instance["security"],
                    "eid_data": security_data_instance["eidData"],
                    "exceptions": security_data_instance["fieldExceptions"],
                    "sequenceNumber": security_data_instance["sequenceNumber"],
                }
                result.update(field_data_instance)
                results.append(result)

    session.stop()
    return results


def get_field_info(fieldname: Optional[str] = None, search: bool = False):
    """If fieldname is None, returns all available fields"""

    svc_type = "//blp/apiflds"
    session = get_blp_api_session(svc_type)
    fld_svc = session.getService(svc_type)

    if search:
        request = fld_svc.createRequest("FieldSearchRequest")
        request.set("searchSpec", fieldname)  # type: ignore
        request.set("returnFieldDocumentation", True)  # type: ignore
    elif fieldname is None or fieldname in ("All", "Static", "RealTime"):  # All Fields
        request = fld_svc.createRequest("FieldListRequest")
        if not fieldname:
            fieldname = "All"
        request.set("fieldType", fieldname)  # type: ignore
        request.set("returnFieldDocumentation", True)  # type: ignore
    else:  # Individual Field
        request = fld_svc.createRequest("FieldInfoRequest")
        request.append("id", fieldname)  # type: ignore
        request.set("returnFieldDocumentation", True)  # type: ignore

    session.sendRequest(request)
    data = await_response(session)

    results = []

    for e in data:
        e_py = e.toPy()
        logger.debug(e_py)
        field_data = e_py["fieldData"]

        for field_data_instance in field_data:
            fid = field_data_instance["id"]
            field_info = (
                field_data_instance["fieldInfo"]
                if "fieldInfo" in field_data_instance
                else field_data_instance["fieldError"]
            )
            ftype = field_info["ftype"] if "ftype" in field_info else None

            result = {"id": fid, "ftype": ftype}
            result.update(field_info)
            results.append(result)

    session.stop()
    return results



import importlib
import inspect
import logging
from functools import lru_cache
from pathlib import Path
from typing import Dict, Iterable, Optional, Tuple, Union

import sqlparse
from pandas import DataFrame

from . import constants, options_parser
from .datamodel.database import IqlDatabaseConnector
from .datamodel.extension import IqlExtension
from .datamodel.querycontainer import IqlQueryContainer
from .datamodel.result import IqlResult
from .utils import extract_subquery_strings

logger = logging.getLogger(__name__)

# Customizations #
# Configure before first use
# Note: Will likely move these to a config file
DB_MODULE: str = "iql.db_connectors.duckdb_connector"  # Change before first db use
DFPREFIX: str = "iqldf"  # Prefix for temporary files
DEFAULT_EXT_DIRECTORY = None  # Sets temp dir for extensions, set before extension use

# Internal Objects
_EXTENSIONS: Dict[Tuple[str, str], IqlExtension] = {}  # register_extension
_DBCONNECTOR: Optional[IqlDatabaseConnector] = None  # Set DB_MODULE before use.
_SUBSTITUTIONS: Dict[str, str] = {}  # Substition map: any occurence in a query of key will be replaced by value
_ALIASES: Dict[str, Union[str, Path]] = {}  # register_alias


def _parameterize_sql_alias(subword, query) -> str:
    if subword not in _ALIASES.keys():
        raise ValueError(f"Unknown alias {subword}")

    query_data = _ALIASES[subword]
    base_query = query_data if isinstance(query_data, str) else query_data.read_text()

    for k, v in options_parser.options_to_list(query).items():
        # convert the parameter to $uppercase, so security => $SECURITY
        newk = f"${k.upper()}"
        base_query = base_query.replace(newk, str(v))

    return base_query


def replace_sql_aliases(query) -> str:
    newquery = query
    for _, subword, outer, _ in extract_subquery_strings(query, ["alias"]):
        sql = _parameterize_sql_alias(subword, outer)

        # aliases might recurse
        sql = replace_sql_aliases(sql)
        logger.debug("Found SQL alias, replacing and parameterizing: %s", outer)

        newquery = newquery.replace(outer, sql)

    return newquery


@lru_cache(maxsize=3)
def parse_sql(query: str):
    return sqlparse.parse(query)


def substitute_replace(query: str, substitutions: Optional[Dict[str, str]]) -> str:
    if substitutions is None:
        all_subs = _SUBSTITUTIONS
    else:  # override any substitutions
        all_subs = _SUBSTITUTIONS.copy()
        all_subs.update(substitutions)

    lastquery = None

    new_query = query
    while new_query != lastquery:
        # Keep changing until all changes and aliases are updated
        # Since an alias can have a substitution, and a substitution can have an alias
        # We just need to repeat

        lastquery = new_query

        logger.debug("Substituting with %s, query=%s", all_subs, new_query)
        for k, v in all_subs.items():
            # logger.debug("%s %s", k, v)
            # TODO: Remove the $, and use exact SUBSTITUTIONS. Makes it easier to use whatever
            # substitution syntax you want.
            # TODO: Consider how to support both JINJA and native
            new_query = new_query.replace(f"${k.upper()}", str(v))
        # Aliases
        new_query = replace_sql_aliases(new_query)

    if new_query != query:
        logger.debug("Updated query %s", new_query)

    return new_query


def execute_debug(
    query: str,
    con: Optional[object] = None,
    substitutions: Optional[Dict[str, str]] = None,
    parameters: Optional[Iterable] = None,
    raw: bool = False,
    context: Optional[Dict[str, object]] = None,
) -> Tuple[bool, Optional[DataFrame], Dict[str, Optional[DataFrame]], Optional[Dict[str, Iterable[IqlResult]]]]:
    """Returns the success (True or False), final query result, and the debug
    results: all the intermediate queries and subqueries."""
    # Connection to database

    # Special case for BQL only.
    if query.strip().startswith("get") or query.strip().startswith("let"):
        # TODO: Handle "from first" bql

        if '"""' in query:
            raise ValueError("Triple quotes are not allowed in query")

        query = query.replace('"', '\\"')
        query = f'''select * from bql(""" {query} """)'''

    # Initialize or Reuse DB Connector
    idc: IqlDatabaseConnector = get_dbconnector()

    if con is None:
        db = idc.create_database()
    else:
        db = idc.create_database_from_con(con=con)

    query = substitute_replace(query, substitutions)

    try:
        # A single query might contain multiple SQL statements. Parse them out and iterate:
        single_df: Optional[DataFrame] = None
        completed_result_map: Dict[str, Optional[DataFrame]] = {}
        intermediate_result_map = {}
        if ";" not in query:
            # Performance optimization for simple queries
            iqc = IqlQueryContainer(query=query, db=db, substitutions=substitutions, context=context)  # type: ignore

            single_df, intermediate_results = iqc.execute(parameters=parameters, raw=raw)
            return (True, single_df, {query: single_df}, {query: intermediate_results})
        else:
            logger.debug("Parsing %.120s", query)

            statements = parse_sql(query)

            for i, statement in enumerate(statements):
                singlequery = statement.value.strip(";")
                iqc = IqlQueryContainer(
                    query=singlequery,
                    db=db,
                    substitutions=substitutions,  # type: ignore
                    context=context,
                )

                # Run each statement, but only keep the results from the last one

                if (
                    i == len(statements) - 1
                ):  # Only pass parameters to last query. This is consistent with how duckdb would run this natively.
                    single_df, completed_results = iqc.execute(parameters=parameters, raw=raw)
                else:
                    single_df, completed_results = iqc.execute(raw=raw)

                completed_result_map[singlequery] = single_df
                intermediate_result_map[singlequery] = completed_results

        return (True, single_df, completed_result_map, intermediate_result_map)

    finally:
        if con is None:  # DB was created here, so close it
            db.close_db()


def execute(
    query: str,
    con: Optional[object] = None,
    substitutions: Optional[Dict[str, str]] = None,
    parameters: Optional[Iterable] = None,
    raw: bool = False,
    context: Optional[Dict[str, object]] = None,
    autocontext: bool = True,
) -> Optional[DataFrame]:
    """Executes the given SQL query. Keyword is only used to run a single subquery without SQL."""

    if autocontext and context is None:
        context = inspect.currentframe().f_back.f_locals  # type: ignore

    success, single_df, completed_results, intermediate_results = execute_debug(
        query, con, substitutions=substitutions, parameters=parameters, raw=raw, context=context
    )

    return single_df


def executedf(*args, **kwargs) -> DataFrame:
    """Helper that enforces that a DataFrame is returned, otherwise ValueError"""

    if kwargs.get("autocontext", True) and kwargs.get("context", None) is None:
        context = inspect.currentframe().f_back.f_locals  # type: ignore
        kwargs["context"] = context

    result_df = execute(*args, **kwargs)

    if result_df is None:
        raise ValueError("Execution returned a None result")
    else:
        return result_df


def get_dbconnector() -> IqlDatabaseConnector:
    global _DBCONNECTOR
    if _DBCONNECTOR is None:
        # Initializes only on first reference
        module = importlib.import_module(DB_MODULE)
        _DBCONNECTOR = module.get_connector()

    return _DBCONNECTOR


def register_extension(e: IqlExtension):
    """* subword means: catchall"""

    _EXTENSIONS[(e.keyword, e.subword)] = e  # type: ignore

    if e.keyword not in constants._KNOWN_EXTENSIONS.keys():
        constants._KNOWN_EXTENSIONS[e.keyword] = e.keyword
        constants._LOADED_EXTENSIONS.append(e.keyword)

    if e.temp_file_directory is None:
        e.temp_file_directory = DEFAULT_EXT_DIRECTORY


def register_alias(subword: str, data: Union[str, Path]):
    """
    Aliases are called via
    alias.aliasname(param1=abc, param2.abc)
    When run, the alias will be replaced, and "$PARAM1" will be replaced with abc.
    You can either register the entire SQL or a path to a file containing the SQL.

    If a Path is given, then it is loaded on each access (not cached).
    """

    logger.debug("Registering alias: %s", subword)
    _ALIASES[subword] = data


def list_extensions() -> Iterable[str]:
    return list(constants._KNOWN_EXTENSIONS.keys())


def get_extension(keyword: str, subword: str) -> IqlExtension:
    """Loads extension on first use"""

    if keyword not in constants._LOADED_EXTENSIONS:
        if keyword not in constants._KNOWN_EXTENSIONS.keys():
            raise ValueError(f"Unknown Extension {keyword}")
        else:
            classname = constants._KNOWN_EXTENSIONS[keyword]
            logger.debug("Loading classname=%s", classname)
            module = importlib.import_module(classname)
            module.register(keyword)
            constants._LOADED_EXTENSIONS.append(keyword)

    words = (keyword, subword)

    ext = _EXTENSIONS.get(words, _EXTENSIONS.get((keyword, "*"), None))
    if ext is not None:
        return ext
    else:
        raise ValueError(f"{keyword}.{subword} is not registered")


def configure(temp_dir: Optional[str] = None):
    """
    Must be called before extensions are initialized (on first use)
    duration_seconds: None (Disabled), -1 (Infinite), int (Seconds)
    file_directory: None (no file cache), string (output directory)
    """
    global DEFAULT_EXT_DIRECTORY
    DEFAULT_EXT_DIRECTORY = temp_dir


def patch_sqlparse():
    # patch sqlparse to allow leading $'s for substitution patterns
    try:
        sqlparse.keywords.SQL_REGEX.insert(9, (r"\$[\w$]*", sqlparse.tokens.Name))
        sqlparse.lexer.Lexer.get_default_instance().set_SQL_REGEX(sqlparse.keywords.SQL_REGEX)
    except Exception:
        logger.exception("Couldn't patch sqlparse to allow a leading $ in table names")



import argparse
import inspect
import logging
import re
from typing import Iterable, Optional

from IPython.core.getipython import get_ipython

from .. import ql

try:
    from jinja2 import Template
except Exception:  # noqa
    # Nothing to do, jinja2 is optional
    # User will get an error when using -j
    pass

from duckdb import DuckDBPyConnection
from IPython.core.magic import (
    Magics,
    cell_magic,
    line_magic,
    magics_class,
    needs_local_scope,
    no_var_expand,
)
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring
from traitlets.config.configurable import Configurable

logger = logging.getLogger(__name__)

CONNECTION: Optional[DuckDBPyConnection] = None


def _get_obj_from_name(name: str) -> Optional[object]:
    ip = get_ipython()
    return ip.ev(name) if ip is not None else None


def execute_iql(
    query_string: str,
    connection: Optional[DuckDBPyConnection],
    export_function: Optional[str] = None,
    query_params: Optional[Iterable] = None,
):
    # Pop back to the calling frame: this is 3 backs in ipython.
    # Why? https://github.com/duckdb/duckdb/pull/12425
    # TODO: Add option to disable
    context = inspect.currentframe().f_back.f_back.f_back.f_locals  # type: ignore
    return ql.execute(query_string, connection, parameters=query_params, context=context, autocontext=False)


@magics_class
class IqlMagic(Magics, Configurable):
    # database connection object
    # created via -d (default), -cn (connection string) or -co (connection object)

    # selected via -t. None = Pandas.
    export_function = None
    shell = None

    def __init__(self, shell):
        Configurable.__init__(self, config=shell.config)
        Magics.__init__(self, shell=shell)

        # Add ourself to the list of module configurable via %config
        self.shell.configurables.append(self)  # type: ignore

    def connect_by_objectname(self, connection_object):
        con: DuckDBPyConnection = _get_obj_from_name(connection_object)  # type: ignore
        if not isinstance(con, DuckDBPyConnection):
            raise ValueError(f"{connection_object} is not a DuckDBPyConnection")
        logger.debug("Using existing connection: %s", connection_object)

        global CONNECTION
        CONNECTION = con

    @no_var_expand
    @needs_local_scope
    @line_magic("iql")
    @cell_magic("iql")
    @line_magic("bql")
    @cell_magic("bql")
    @magic_arguments()
    @argument(
        "-co",
        "--connection_object",
        help="Connect to a database using the connection object",
        nargs=1,
        type=str,
        action="store",
    )
    @argument(
        "-t",
        "--type",
        help="Set the default output type",
        nargs=1,
        type=str,
        action="store",
    )
    @argument(
        "-o",
        "--output",
        help="Write the output to the specified variable",
        nargs=1,
        type=str,
        action="store",
    )
    @argument(
        "-q",
        "--quiet",
        help="Don't return an object, similar to %%capture",
        action="store_true",
    )
    @argument("-j", "--jinja2", help="Apply Jinja2 Template", action="store_true")
    @argument(
        "-s",
        "--substitute",
        nargs=1,
        type=str,
        action="store",
        help="Substitute Variables using a Dictionary",
    )
    @argument(
        "-p",
        "--param",
        dest="queryparams",
        help="Params from user_ns",
        action="append",
    )
    @argument(
        "--ext",
        dest="extension",
        nargs=1,
        type=str,
        action="store",
        help="Wrap all commands in extension",
    )
    @argument("rest", nargs=argparse.REMAINDER)
    def execute(self, line: str = "", cell: str = "", local_ns=None):  # noqa: C901
        global CONNECTION

        args = parse_argstring(self.execute, line)
        # Grab rest of line
        rest = " ".join(args.rest)
        query = f"{rest}\n{cell}".strip()

        logger.debug("Query = %s", query)

        if args.substitute:
            substitutions = self.shell.user_ns[args.substitute[0]]
            for k, v in substitutions.items():
                query = query.replace(k, str(v))

        if args.jinja2:  # Replace any {var}'s with the string values
            query = Template(query).render(self.shell.user_ns)  # type: ignore

        # Strip any DBT "{{ref('table')}}"
        query = re.sub(r"{{\s*ref\(['\"](.*?)['\"]\)\s*}}", r"\1", query)

        if args.connection_object:
            self.connect_by_objectname(args.connection_object[0])

        if args.queryparams:
            query_params = [self.shell.user_ns[p] for p in args.queryparams]
        else:
            query_params = None

        if query is None or len(query) == 0:
            logger.debug("Nothing to execute")
            return

        # DBT Jinja References
        # Replace any jinja tables / refs {{ref ('nc_sapi')}} for instance
        query = re.sub(r"\{\{\s*ref\s*\(\s*'([^']+)'\s*\)\s*\}\}", r"\1", query)

        if args.extension:
            extension = args.extension[0]
            logger.debug("Wrapping query with extension %s", extension)
            query = f'''{extension}(""" \n {query} \n""")'''

        o = execute_iql(
            query_string=query,
            connection=CONNECTION,
            export_function=self.export_function,
            query_params=query_params,
        )

        if args.type:
            if args.type[0] == "markdown":
                print(o.to_markdown())  # type: ignore  # noqa
                return None

        if args.output:
            self.shell.user_ns[args.output[0]] = o  # type: ignore

        if args.quiet:
            return None
        else:
            return o


def init_completers(ip):
    try:
        logger.info("Initializating autocompleter")
        from .autocompletion_v2 import init_completer

        init_completer(ipython=ip)
    except Exception:
        logger.debug("Unable to initialize autocompletion_v2. iPython 8.6.0+ is required. Trying v1 completer")
        try:
            from .autocompletion_v2 import init_completer

            init_completer(ipython=ip)
        except Exception:
            logger.debug("Unable to initialize autocompletion_v1")



import logging
import os
import time
from dataclasses import dataclass

import openai
import pandas as pd

from ..datamodel import cache
from ..datamodel.cache import iql_cache
from ..datamodel.extension import IqlExtension
from ..datamodel.subquery import SubQuery
from ..ql import register_extension

_CREDENTIAL = None

logger = logging.getLogger(__name__)


def get_result(key: str, assistant_id: str, prompt: str):
    openai.api_key = key
    client = openai.OpenAI(api_key=key)
    thread = client.beta.threads.create()

    openai.beta.threads.messages.create(thread_id=thread.id, role="user", content=prompt)

    run = openai.beta.threads.runs.create(thread_id=thread.id, assistant_id=assistant_id)

    while True:
        run_status = openai.beta.threads.runs.retrieve(thread_id=thread.id, run_id=run.id)
        if run_status.status == "completed":
            break
        time.sleep(1)

    messages = openai.beta.threads.messages.list(thread_id=thread.id)
    return messages.data[0].content[0].text.value


@dataclass
class OpenAiExtension(IqlExtension):
    """
    DuckDB does support native reading of certain file types.
    This is fine for XLSX files, but some file types need different engines, such as "xls", that duckdb doesn't support
    """

    param_replace_text = False
    keyword: str
    is_async: bool = False

    @iql_cache
    def executeimpl(self, sq: SubQuery) -> pd.DataFrame:
        if "openai_key" in sq.options:
            openai_key: str = sq.options.get("openai_key")  # type: ignore
        else:
            openai_key: str = os.environ.get("OPENAI_KEY", None)  # type: ignore

        if "openai_assistant_id" in sq.options:
            assistant_id: str = sq.options.get("openai_assistant_id")  # type: ignore
        else:
            assistant_id: str = os.environ.get("OPENAI_ASSISTANT_ID", None)  # type: ignore

        if not openai_key:
            raise ValueError("OPENAI_KEY not set")

        prompt: str = sq.options.get("prompt")  # type: ignore
        context: str = sq.options.get("context")  # type: ignore

        result = get_result(key=openai_key, assistant_id=assistant_id, prompt=prompt)
        return pd.DataFrame({"context": [context], "result": [result]})


def register(keyword: str):
    extension = OpenAiExtension(keyword=keyword)
    extension.cache = cache.MemoryAndFileCache(max_age=3600 * 24, return_pyarrow_table=False)
    register_extension(extension)



import logging
from dataclasses import dataclass
from html import unescape
from io import BytesIO

import pandas as pd
from azure.core.credentials import AccessToken, TokenCredential
from azure.storage.blob import BlobServiceClient

from iql import IqlExtension, SubQuery, iql_cache, register_extension

_CREDENTIAL = None

logger = logging.getLogger(__name__)


class StaticTokenCredential(TokenCredential):
    def __init__(self, access_token: AccessToken):
        self._access_token = access_token

    def get_token(self, *scopes, **kwargs) -> AccessToken:
        return self._access_token


@dataclass
class AzureBlobExtension(IqlExtension):
    """
    DuckDB does support native reading of certain file types.
    This is fine for XLSX files, but some file types need different engines, such as "xls", that duckdb doesn't support
    """

    param_replace_text = False
    keyword: str
    is_async: bool = False

    @iql_cache
    def executeimpl(self, sq: SubQuery) -> pd.DataFrame:
        blob_url: str = sq.options["url"]  # type: ignore

        logger.info("Reading %s", blob_url)
        engine: str | None = sq.options.get("engine", None)  # type: ignore
        account_url = blob_url.split("/")[2]
        # account_name = blob_url.split(".")[0]
        filename = unescape(blob_url.split("/")[-1])
        container = blob_url.split("/")[3]

        coerce_str = sq.options.get(
            "coerce_str", True
        )  # by default, cooerce everything to string. Avoids parquet export errors.

        if not filename.endswith(".xls") and not filename.endswith(".xlsx"):
            raise ValueError("Only .xls or .xlsx is supported")

        if filename.endswith(".xls") and engine is None:
            engine = "calamine"

        blob_service_client = BlobServiceClient(account_url, credential=_CREDENTIAL)
        blob_client = blob_service_client.get_blob_client(container=container, blob=filename)

        data = blob_client.download_blob().readall()
        data_io = BytesIO(data)

        # The astype(str) is needed because of the mixed data types in the columns, which will break parquet export
        if engine is not None:
            result_df = pd.read_excel(data_io, engine=engine)  # type: ignore
        else:
            result_df = pd.read_excel(data_io)

        if coerce_str:
            result_df = result_df.astype(str)

        result_df["blob_url"] = blob_url
        return result_df


def register(keyword: str):
    extension = AzureBlobExtension(keyword=keyword, subword="excel")
    # extension.cache = cache.MemoryAndFileCache(max_age=3600 * 24, return_pyarrow_table=False)
    register_extension(extension)



"""This module wraps the Bloomberg BQL API to provide
consistent return results"""

import logging
import re
import time
from abc import abstractmethod
from collections import Counter, defaultdict
from functools import partial
from typing import Dict, Iterable, Optional, Union

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from pandas import DataFrame

from ...datamodel.cache import SqCache
from ...datamodel.subquery import harmonize_and_concat_arrow_tables

# Pandas concat allows differing schemas. Pyarrow concat requires harmonizing first
USE_ARROW_DTYPES = False  # experimental

_MAX_CONCURRENT = 128

logger = logging.getLogger(__name__)

_merge_value_col = True


class BaseBqlQuery:
    STATUS_NOTRUN = "NOTRUN"
    STATUS_COMPLETED = "COMPLETED"
    STATUS_FAILURE = "FAILURE"

    execution_status: str = STATUS_NOTRUN
    exception_msg: Optional[str] = None
    _data: Optional[Iterable[Dict[str, object]]] = None
    _df: Optional[DataFrame] = None
    _output_filename: Optional[str] = None  # Used for large queries where we don't want to hold it all in memory

    params = None

    @abstractmethod
    def get_fields(self) -> Iterable[str]:
        pass

    @abstractmethod
    def to_bql_query(self) -> str:
        pass

    def execute(self) -> bool:
        """Executes the query and, if successful, sets the query._data"""
        self._populate_data()
        return self.execution_status == self.STATUS_COMPLETED

        # All exceptions should be well handled
        # except Exception as e:
        #    logger.exception("execute")
        #    self.execution_status = self.STATUS_FAILURE
        #    self.exception_msg = str(e)
        #    return False

    def to_data(self) -> Union[DataFrame, Iterable[Dict[str, object]]]:
        """Internal representation"""
        if self._data is not None:
            return self._data
        else:
            success = self.execute()
            if success and self._data is not None:
                return self._data
            elif success and self._df is not None:
                return self._df
            else:
                raise ValueError("Failure executing BQL query")

    def to_df(self) -> DataFrame:
        if self._df is None:
            self._df = DataFrame(self._data)
        return self._df

    def _populate_data(self):
        execute_bql_str_list_async_q([self])


def list_to_str(values: Iterable, quote: bool = False, delimiter: str = ", \n") -> str:
    """Helper to convert list of values to a comma delimited list. Used for BQL functions.
    Equity lists should be quoted"""

    if quote:
        return delimiter.join(f"'{val}'" for val in values)
    else:
        return delimiter.join(values)


def security_to_finalstr(security: Union[str, Iterable]) -> str:
    if isinstance(security, str) and ("(" in security or "[" in security or "$" in security):
        final_security_str = security
    elif isinstance(security, str):
        if "'" not in security:
            security = "'" + security + "'"
        final_security_str = "[" + security + "]"
    else:  # Iterable
        final_security_str = list_to_str(security, True)
        final_security_str = "[" + final_security_str + "]"

    return final_security_str


def construct_bql_query(
    field_str: Union[str, Iterable],
    security: Union[Iterable[str], str],
    with_params: Optional[str] = None,
    let_vars: Optional[str] = None,
) -> str:
    """Simple wrapper to construct a valid BQL query string from an
    already comma delimited list of fields and quoted securities"""

    # Better to use BQLQuery, but leaving this for now.

    if isinstance(field_str, str):
        field_str = [field_str]

    field_str = list_to_str(field_str, False)

    final_security_str = security_to_finalstr(security)

    request_str = ""
    if let_vars is not None:
        request_str += "let (" + let_vars + ")"

    request_str += "get("
    request_str += field_str

    request_str += ") for (" + final_security_str + ")"

    # default WITH clause
    # TODO: Replace this with something metadata driven, since it may not be applicable in all cases
    # request_str += "\nwith (fill=prev)"

    if with_params is not None:
        request_str += " with (" + with_params + ")"

    if "preferences" not in request_str:
        request_str += "\npreferences (addcols=all)"

    return request_str


def bql_exception_to_str(e) -> object:
    try:
        e = e[1]
        long_error = {
            "message": str(e.exception_messages),
            "request_id": str(e._request_id),
            "details": str(e.internal_messages),
        }
        return long_error
    except Exception:
        return str(e)


def error_callback(o, errorlist=None):
    if errorlist is None:
        logger.warning(o)
        return

    msg = bql_exception_to_str(o)
    errorlist.append(msg)


def execute_bql_str_list_async_q(
    queries_input: Iterable[BaseBqlQuery],
    suppress_warning_log: bool = False,
    max_queries: int = _MAX_CONCURRENT,
    allow_async: bool = True,
    cache: SqCache | None = None,
):
    """Note: Not multi-process safe due to underlying BQL APIs
    and (presumably) its use a Singleton for bqapi requests.
    Either batch requests through a single requester thread, or
    distribute workload across requesters.
    """
    logger.debug("Checking to see if any are already cached")

    for q in queries_input:
        qstr = q.to_bql_query()
        data = None if cache is None else cache.get(qstr)
        if data is not None:
            logger.debug("Found in cache: %s", qstr)

            if isinstance(data, DataFrame):
                q._df = data
            else:
                q._data = data  # type: ignore
            q.execution_status = q.STATUS_COMPLETED

    queries_not_cached = [q for q in queries_input if q._data is None and q._df is None]

    # Max Queries indicates the max per batch: chop the queries
    # into smaller groups of max_queries size and execute each serially

    if len(queries_not_cached) == 0:
        logger.debug("No queries to run")
        return

    query_groups = [
        queries_not_cached[i * max_queries : (i + 1) * max_queries]
        for i in range((len(queries_not_cached) + max_queries - 1) // max_queries)
    ]

    start = time.time()

    count = 1
    for query_group in query_groups:
        logger.debug(
            "Executing %s of %s query batches with %s queries",
            count,
            len(query_groups),
            len(query_group),
        )
        count = count + 1

        # t1 = threading.Thread(target=asyncio.run, args=(_execute_bql_str_list_async_callbacks(query_group, suppress_warning_log),))
        # t1.start()
        # t1.join

        logger.debug("Done async with callbacks")
        _execute_bql_str_list_async_orig(query_group, suppress_warning_log, allow_async=allow_async)

    end = time.time()
    logger.debug("Elapsed time running queries: %s seconds", round(end - start))

    for q in queries_not_cached:
        if q._data is not None and cache is not None:
            logger.debug("Saving to cache")

            cache.save(q.to_bql_query(), q.to_df(), cost=None)

    return


def is_bquant():
    return False  # os.environ.get("BQUANT_USERNAME") is not None


def _execute_bql_str_list_async_orig(
    queries: Iterable[BaseBqlQuery],
    suppress_warning_log: bool = False,
    allow_async: bool = True,
):
    from . import (
        bql_service,
    )

    with bql_service.BQL_LOCK:
        query_strings = [query.to_bql_query() for query in queries]

        bq_service = bql_service.get_bqservice()

        # TODO: Out of order errors are not handled here. Not an issue in many cases
        # but is still very likely.
        errorlist: Iterable[str] = []

        try:
            if is_bquant() and allow_async:
                logger.debug("Using submit fetch many")

                gen = bq_service._submit_fetch_many(
                    query_strings,
                    on_request_error=partial(error_callback, errorlist=errorlist),
                    num_retries=1,
                )
            else:
                gen = bq_service.execute_many(
                    query_strings,
                    on_request_error=partial(error_callback, errorlist=errorlist),
                )
        except Exception as e:
            logger.exception(bql_exception_to_str(e))
            raise ValueError(f"Error submitting {query_strings}") from e

        error_index = 0

        logger.info("Waiting for and processing BQL results")

        for r, q in zip(gen, queries, strict=False):
            try:
                logger.info("Processing request")
                if r is None:
                    q.execution_status = q.STATUS_FAILURE
                    logger.warning("Error executing: %s", q.to_bql_query())
                    q.exception_msg = errorlist[error_index]
                    error_index = error_index + 1
                else:
                    if q._output_filename is not None:
                        result_table: pa.Table = _to_data_arrow(r, return_pandas=False)  # type: ignore
                        logger.info("Writing result to parquet: %s", q._output_filename)
                        pq.write_table(result_table, q._output_filename)
                    else:
                        result_table: DataFrame = _to_data_arrow(r, return_pandas=True)
                        q._df = result_table
                        q._data = result_table  # type: ignore

                    q.execution_status = q.STATUS_COMPLETED
                logger.info("Done processing request")

            except Exception as e:
                logger.exception("Query error")
                q._data = None
                q.execution_status = q.STATUS_FAILURE
                q.exception_msg = bql_exception_to_str(e)  # type: ignore
        logger.debug("Done executing")


def _process_single_item_response_arrow(response, results: Iterable[pa.Table]):  # noqa: C901
    d = response._SingleItemResponse__result_dict

    if len(d["responseExceptions"]) > 0:
        for r in d["responseExceptions"]:
            if r["type"] != "BENIGN":
                raise ValueError(f"{response.name=} {d['responseExceptions']=}")
            else:
                logger.debug("Benign BQL Exception: %s", r)
    try:
        logger.debug("Arrow processing %s", d["name"])  # {d}

        ids = d["idColumn"]
        vcol = ids
        dcol = d["valuesColumn"]
        acol = d["secondaryColumns"]

        cols = [vcol, dcol, *acol]
        data = {col["name"].lower(): col["values"] for col in cols}
        data["name"] = d["name"]

        if dcol["name"].lower() != "value":
            # if the data/value field is not named "value", rename it
            logger.debug("Changing %s to value", dcol["name"])
            dcol["name"] = "value"

        foundcols = set()
        foundcols.add("name")

        def replace_cols(c, type: str) -> str:
            colname = c["name"].lower()
            i = 0
            newcol = colname
            while newcol in foundcols:
                i += 1
                newcol = f"{colname}_{i}"
            if newcol != colname:
                logger.debug("Rewrote %s to %s", colname, newcol)

            if newcol == "value":
                value_field = f"value_{type}"
                newcol = value_field

            foundcols.add(newcol)
            return newcol

        fields = [
            (
                pa.field(replace_cols(col, "str"), pa.string())
                if col["type"] == "STRING" or col["type"] == "ENUM"
                else (
                    pa.field(replace_cols(col, "date"), pa.timestamp("ms"))
                    if col["type"] == "DATE"
                    else (
                        pa.field(replace_cols(col, "float"), pa.float64())
                        if col["type"] == "FLOAT" or col["type"] == "DOUBLE"
                        else pa.field(replace_cols(col, "int"), pa.int64())
                    )
                )
            )
            for col in cols
        ]

        arrays = [
            # pa.array([str(val) for val in col["values"]])
            # if col["type"] == "STRING"
            # else
            pa.array(
                [
                    (
                        val
                        if val != "NaN"
                        and val != "NaT"
                        and not pd.isna(val)
                        and val != "Infinity"
                        and val != "-Infinity"
                        else None
                    )
                    for val in col["values"]
                ]
            )
            for col in cols
        ]

        for i, c in enumerate(cols):
            if c["type"] == "DATE":
                arrays[i] = arrays[i].cast(pa.timestamp("ms", tz="UTC"))

        name = d["name"].lower()
        if name[0] == "#":  # strip the leading # from aliases
            name = name[1:]

        namearray = [name] * len(arrays[0])

        arrays.append(pa.array(namearray))
        fields.append(pa.field("name", pa.string()))

        # make a copy of the value field, so we'll have value and value_int, for instance
        if _merge_value_col:
            value_idx = 1
            f = fields[value_idx]
            fields.append(pa.field("value", f.type))
            arrays.append(arrays[value_idx])

        table = pa.Table.from_arrays(arrays, schema=pa.schema(fields))

        results.append(table)  # TODO: Research switching to types_mapper=pd.ArrowDtype
    except Exception as e:
        logger.exception(d)
        raise ValueError(f"Error processing SIR {d['name']}") from e


def _process_response_arrow(response, results):  # type: ignore
    for res in response._sirs:
        if not hasattr(res, "_SingleItemResponse__result_dict"):
            raise ValueError("Response element does not have a _SingleItemResponse__result_dict")
        _process_single_item_response_arrow(res, results)  # type: ignore

    # This small set = None avoids a circular reference that was pinning memory consumption for very large queries.
    response._sirs = None


def _to_data_nonarrow_execute_response(response) -> pa.Table:  # type: ignore
    """Converts BQL Response to a List of Dicts, where each dict contains Col: Value"""
    result_table: Iterable[pa.Table] = []

    _process_response_arrow(response, result_table)

    logger.debug("Done processing, %s tables", len(result_table))

    final_table = harmonize_and_concat_arrow_tables(result_table)

    # if as_arrow:
    return final_table
    # else:
    #     if USE_ARROW_DTYPES:
    #         return final_table.to_pandas(types_mapper=pd.ArrowDtype)
    #     else:  # numpy dtypes / default
    #         return final_table.to_pandas()


def handle_duplicate_cols(table: pa.Table) -> pa.Table:
    if len(set(table.schema.names)) < len(table.schema.names):
        names = table.schema.names
        counts = Counter()
        new_names = []
        for name in names:
            if counts[name]:
                new_names.append(f"{name}_{counts[name]}")
            else:
                new_names.append(name)
            counts[name] += 1
        table = table.rename_columns(new_names)

    return table


def fix_suffixes(all_cols):
    col_renamed = defaultdict(list)
    new_cols = {}
    for origcol in all_cols:
        col = origcol.lower()
        match = re.fullmatch(r"(.*?):(\d+)", col)
        if match:
            before, after = match.groups()
            col_renamed[before].append(int(after))
        else:
            before = col
            col_renamed[before].append(None)

        if before in new_cols.values():
            index = len(col_renamed[before]) - 1
            new_cols[origcol] = f"{before}_{index}"
        else:
            new_cols[origcol] = before
    return new_cols


def transform_table(item: "bqle.request.execute_response_handler._BQLItem") -> pa.Table:  # type: ignore  # noqa: F821, C901
    if item.table is None:
        return None

    table = item.table

    table = handle_duplicate_cols(table)

    if len(item.value_column_index) != 1:
        raise ValueError("Unexpected {item.value_column_index=}")

    coldesc = item.columns[item.value_column_index[0]]  # next(cd for cd in item.columns if cd.name=="VALUE")

    if coldesc.data_type.name == "DOUBLE":
        copy_col = "value_float"
    elif coldesc.data_type.name == "DATE":
        copy_col = "value_date"
    elif coldesc.data_type.name == "STRING":
        copy_col = "value_string"
    else:
        logger.debug("Other data type: %s", coldesc.data_type)
        suffix = coldesc.data_type.name.lower()
        copy_col = f"value_{suffix}"

    # Make every date a timestamp('ms')
    for i, field in enumerate(table.schema):
        if pa.types.is_date32(field.type):
            table = table.set_column(i, field.name, table.column(i).cast(pa.timestamp("ms")))

    if copy_col is not None:
        table = table.append_column(copy_col, table[coldesc.name])

    if coldesc.name != "VALUE":
        table = table.append_column("value", table[coldesc.name])

    column_newnames = fix_suffixes(table.schema.names)  ## {name: fix_col(name) for name in table.schema.names}
    # print(column_newnames)
    if "NAME" in column_newnames:
        # rename NAME column to name_1, if it exists
        i = 1
        while f"name_{i}" in column_newnames.values():
            i += 1

        column_newnames["NAME"] = f"name_{i}"

    table = table.rename_columns(list(column_newnames.values()))
    colname = (item.name[1:] if item.name.startswith("#") else item.name).lower()

    table = table.append_column("name", pa.array([colname] * len(table)))

    return table


def _to_data_arrow_execute_response(response) -> pa.Table:
    failed = False
    tables = []
    for sir in response._sirs:
        if sir._bql_item.table is not None:
            transformed_table = transform_table(sir._bql_item)
            tables.append(transformed_table)
        else:
            tables.append(pa.table({"name": [], "id": []}))

        for e in sir._bql_item.errors:
            if e["code"].startswith("BENIGN"):
                logger.debug("Item %s failed with %s, BENIGN: continuing", sir._bql_item.name, sir._bql_item.errors)
            else:
                logger.info("Item %s failed with %s, FAILED", sir._bql_item.name, sir._bql_item.errors)

                failed = True

    # failed = any(sir._bql_item.failed for sir in response._sirs)

    # tables = [transform_table(sir._bql_item) for sir in response._sirs]

    response._sirs = None  # Might be unnecessary, but was used to avoid circular references in the pre-Arrow sirs

    if failed:
        raise ValueError("Query failed, {tables=}")

    return harmonize_and_concat_arrow_tables(tables)


def _process_arrow_single_item_response_arrow(
    sir: "bqle.common.arrow_execute_response.ArrowExecuteSingleItemResponse",  # type: ignore # noqa: C901, F821
    results: Iterable[pa.Table],
):
    if len(sir._bql_item.errors) > 0:
        raise ValueError("Errors in _bql_item: sir._bql_item.errors=}")


def _to_data_arrow(response, return_pandas: bool = True) -> DataFrame:  # type: ignore
    if type(response).__name__ == "ArrowExecuteResponse":
        table = _to_data_arrow_execute_response(response)
    else:
        table = _to_data_nonarrow_execute_response(response)

    if return_pandas:
        return table.to_pandas()
    else:
        return table



import logging
from dataclasses import dataclass

import pandas as pd

from ..datamodel.extension import IqlExtension
from ..datamodel.subquery import SubQuery
from ..ql import executedf, register_extension

logger = logging.getLogger(__name__)


@dataclass
class NestedQueryExtension(IqlExtension):
    keyword: str

    def executeimpl(self, sq: SubQuery) -> pd.DataFrame:
        """Executes the query in the same database as this query
        Example: select * from nestiql("alias.stats();select 1");

        """
        query = sq.get_query()

        if query is None:
            raise ValueError("Query must be passed")

        if sq.options.get("nested_connection", True):
            con = sq.iqc.db._connection
        else:
            con = None

        result = executedf(query, con=con)
        return result


def register(keyword: str):
    extension = NestedQueryExtension(keyword=keyword)
    register_extension(extension)



import logging
import re
from typing import Iterable, Optional, Tuple

from .constants import _KNOWN_EXTENSIONS

logger = logging.getLogger(__name__)


def _find_closing_paren(text: str, start: int) -> int:
    # TODO: Handle escaping
    paren_depth = 1
    quote_stack = []
    for i in range(start, len(text)):
        c = text[i]
        if c == "(" and len(quote_stack) == 0:
            paren_depth += 1
        elif c == "'" or c == '"':
            if len(quote_stack) == 0:
                quote_stack.append(c)
            elif quote_stack[len(quote_stack) - 1] == c:
                quote_stack.remove(c)
        elif c == ")" and len(quote_stack) == 0:
            paren_depth -= 1

        if paren_depth == 0:
            end = i
            return end
    raise ValueError("Never found closing parenthesis, probably unbalanced query")


def extract_subquery_strings(query, keywords: Optional[Iterable[str]] = None) -> Iterable[Tuple[str, str, str, str]]:
    """Finds the subqueries start with keyword, along with matching end paren
    keyword(....)
    keyword.subword(....)
    """
    _keyword_list = get_keyword_list(keywords)

    _kpat = re.compile(rf"(?s)({_keyword_list})(\.(\w+))?\s*\(")

    results = []

    last_paren_end = -1

    for m in re.finditer(_kpat, query):
        logger.debug("finditer: %s", m.group(0))
        keyword = m.group(1)
        subword = m.group(3)
        paren_start = m.end()  # just after the last paren

        if paren_start < last_paren_end:
            logger.debug("Skipping embedded subquery: %s", m.group(0))

            continue

        paren_end = _find_closing_paren(query, paren_start)

        outer = query[m.start() : paren_end + 1]
        inner = query[paren_start : paren_end + 1]
        logger.debug(
            "Extracted subquery: %s.%s at %s:%s",
            keyword,
            subword,
            paren_start,
            paren_end,
        )

        results.append((keyword, subword, outer, inner))
        last_paren_end = paren_end

    return results


def get_keyword_list(keywords: Optional[Iterable[str]] = None) -> str:
    if keywords is None:
        keywords = list(_KNOWN_EXTENSIONS.keys())
    _keyword_list = "|".join((k if isinstance(k, str) else k[0] for k in keywords))
    return _keyword_list



import logging
from typing import List

from pandas import DataFrame

from .. import constants, ql
from . import iql_magic

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

sql_expects_tablename = [
    "UNION",
    "UNION ALL",
    "UNION ALL BY NAME",
    "UNION BY NAME",
    "JOIN",
    "INNER JOIN",
    "LEFT JOIN",
    "RIGHT JOIN",
    "FULL JOIN",
    "LEFT OUTER JOIN",
    "RIGHT OUTER JOIN",
    "FROM",
    "INTO",
]

sql_phrases = [
    "PRAGMA",
    "SELECT",
    "WHERE",
    "GROUP BY",
    "ORDER BY",
    "LIMIT",
    "INSERT",
    "UPDATE",
    "DELETE",
    "ALTER",
    "DROP",
    "TRUNCATE",
    "TABLE",
    "DATABASE",
    "INDEX",
    "VIEW",
    "FUNCTION",
    "PROCEDURE",
    "TRIGGER",
    "AND",
    "OR",
    "NOT",
    "BETWEEN",
    "LIKE",
    "IN",
    "NULL",
    "IS",
    "EXISTS",
    "COUNT",
    "SUM",
    "MIN",
    "MAX",
    "AVG",
    "DISTINCT",
    "AS",
    "CREATE TABLE",
    "CREATE OR REPLACE TABLE",
    "CREATE TABLE IF NOT EXISTS",
    "CREATE VIEW",
]

pragma_phrases = [
    "PRAGMA version",
    "PRAGMA database_list",
    "PRAGMA database_size",
    "PRAGMA show_tables",
    "PRAGMA show_tables_expanded",
    "PRAGMA table_info('",
    "PRAGMA functions",
    "PRAGMA collations",
    "PRAGMA enable_progress_bar",
    "PRAGMA disable_progress_bar",
    "PRAGMA enable_profiling",
    "PRAGMA disable_profiling",
    "PRAGMA disable_optimizer",
    "PRAGMA enable_optimizer",
    "PRAGMA enable_verification",
    "PRAGMA disable_verification",
    "PRAGMA verify_parallelism",
    "PRAGMA disable_verify_parallelism",
    "PRAGMA force_index_join",
    "PRAGMA force_checkpoint",
]

function_query = """
with f_inner as (select 
    function_name, 
    case when len(parameters)=0 then '' else list_reduce(parameters, (x,i) -> x||', ' || i) end param_str,
    '(' || param_str || ')' as param_ex
 from duckdb_functions()
)
select function_name || first(param_ex) as function_full
from f_inner
group by function_name
order by function_full

"""


def get_function_names(ipython, token) -> List[str]:
    try:
        if iql_magic.connection is not None:
            functions = iql_magic.connection.sql(function_query)
            function_list = list(functions.df()["function_full"])
            extensions = list(constants._KNOWN_EXTENSIONS.keys())
            all_functions = function_list + extensions
            filtered_functions = [c for c in all_functions if c.startswith(token)]
            logger.debug("Getting function names: %s", filtered_functions)

            return filtered_functions
    except Exception:
        logger.exception("Unable to get function names")

    return []


def get_table_names(ipython) -> List[str]:
    try:
        user_keys = [k for k, v in ipython.user_ns.items() if isinstance(v, DataFrame)]

        if iql_magic.connection is not None:
            tables = iql_magic.connection.sql("show tables")
            logger.debug("Got tables: %s", tables)

            if tables is None:
                return user_keys
            else:
                return list(tables.df()["name"]) + user_keys
        else:
            logger.debug(user_keys)
            return user_keys
    except Exception:
        logger.exception("Unable to get table names")
        return []


def get_column_names_for_table(ipython, tablename: str) -> List[str]:
    try:
        # check if an object
        # o = ipython.ev(tablename)
        o = ipython.user_ns.get(tablename)
        if o is None and iql_magic.connection is not None:
            columns = list(ql.executedf(f"describe select * from {tablename}")["column_name"].to_numpy())
            if columns is None:
                logger.debug("None columns")
                return []
            else:
                logger.debug("Column names: %s", columns)
                return columns
        elif o is not None:
            if isinstance(o, DataFrame):
                return list(o.columns)
            else:
                logger.debug("%s in namespace, but not a DataFrame %s", tablename, type(o))
                return []
        else:
            return []
    except Exception:
        logger.exception("Unable to get column names")
        return []


def get_column_names(ipython, tablename: str) -> List[str]:
    if tablename == "alias":
        aliases = [c + "()" for c in ql._ALIASES.keys()]
        return aliases
    elif tablename.startswith("$"):
        return get_column_names_for_table(ipython=ipython, tablename=tablename)
    else:
        return get_column_names_for_table(ipython=ipython, tablename=tablename)



from __future__ import annotations

import logging
import re
from typing import Dict, Iterable, Optional, Tuple

import sqlparse
from pandas import DataFrame

from .. import ql
from ..utils import extract_subquery_strings
from . import subquery
from .database import IqlDatabase
from .result import IqlResult, IqlResultData

logger = logging.getLogger(__name__)


class IqlQueryContainer:
    # This is used so we can run the bql_queries as an async batch, separate from processing the results.
    orig_query: str
    query: str
    subqueries: Iterable[subquery.SubQuery]
    substitutions: Dict[str, str]
    db: IqlDatabase
    context: Optional[Dict[str, object]]

    def extract_subqueries(self):
        subqueries: Iterable[subquery.SubQuery] = []

        i = 0

        replacements: Iterable[Tuple[str, Optional[str]]] = []

        if "--" in self.orig_query:
            query = sqlparse.format(self.orig_query, strip_comments=True).strip()
        else:
            query = self.orig_query

        for keyword, subword, outer, _ in extract_subquery_strings(query):
            i += 1
            name = f"{ql.DFPREFIX}_{i}"

            extension = ql.get_extension(keyword, subword)
            sqs: Iterable[subquery.SubQuery] = extension.create_subqueries(query=outer, name=name, iqc=self)
            subqueries.extend(sqs)

            if extension.use_path_replacement():
                names = [sq.extension.get_path_replacement(sq) for sq in sqs]
            else:
                names = [sq.name for sq in sqs]

            if len(names) != 1:
                raise ValueError(f"Unexpected length of names: {len(names)}")

            result = names[0]  # holdover from using unions, instead of subquerygroups.

            replacements.append((outer, result))

        for old, new in replacements:
            logger.debug("Replacing %s with %s", old, new)
            query = query.replace(old, new, 1)  # type: ignore

        logger.debug("After replacements %s", query)
        # if only one function without a beginning subquery
        if len(replacements) > 0:
            values = [value for key, value in replacements if value is not None]
            if re.match(rf"(?s)\s*({'|'.join(values)}).*", query) is not None:
                logger.debug("Starts with replacement: %s", values)
                query = f"select * from {query}"

        self.query = query

        return subqueries

    def __init__(
        self, query: str, db: IqlDatabase, substitutions: Optional[Dict] = None, context: Optional[Dict] = None
    ):
        if substitutions is None:
            substitutions = {}

        self.orig_query = query
        self.query = query
        self.db = db
        self.substitutions = substitutions
        self.context = context

        self.subqueries = self.extract_subqueries()

        # Sanity checks. These don't take into account escaping or quoting, so they're just warnings.
        if query.count("(") != query.count(")"):
            logger.warning("Left/Right parens aren't balanced")

        if query.count("'") % 2 != 0:
            logger.warning("Uneven number of single quotes")

        if query.count('"') % 2 != 0:
            logger.warning("Uneven number of double quotes")

    def get_subqueries_by_extension(self, keyword: str, subword: str) -> Iterable[subquery.SubQuery]:
        results: Iterable[subquery.SubQuery] = []
        results = [s for s in self.subqueries if s.extension.keyword == keyword and s.extension.subword == subword]

        return results

    def execute(  # noqa: C901
        self, raw: bool, parameters: Optional[Iterable] = None
    ) -> Tuple[Optional[DataFrame], Iterable[IqlResult]]:  # noqa: C901
        """Returns the final df, plus the intermediate subquery dataframes.
        If the query is not a Select, returns a None"""
        # Execute the subqueries

        completed_results: Iterable[IqlResult] = []

        for (
            keyword,
            subword,
        ), e in ql._EXTENSIONS.copy().items():  # copy to prevent dictionary changed size during iteration error
            sqs: Iterable[subquery.SubQuery] = self.get_subqueries_by_extension(keyword, subword)  # type: ignore
            sqs = list(sqs)
            if len(sqs) == 0:
                continue

            try:
                e_results = e.execute_batch(sqs)  # type: ignore
            except Exception as e:
                logger.debug("Failed to execute batch %s %s: %s", keyword, subword, self.query)
                raise e
            completed_results += e_results

        # Optimization: If there's just a single subquery, and it's a DF, just return that
        # No need to pass it through duckdb
        # This may have typing implications though.
        if self.query == "select * from iqldf_1" and len(completed_results) == 1:
            # shortcut for a simple case:
            cr1 = completed_results[0].native()
            if isinstance(cr1, DataFrame):
                return (cr1, completed_results)

        result = self.db.execute_query(
            query=self.query, completed_results=completed_results, parameters=parameters, raw=raw, context=self.context
        )  # type: ignore

        if result is None:
            df = None
        elif raw:
            df = result.native()
        else:
            df = result.df()

        # add the final result to the completions
        final_result = IqlResultData(name="final", query=self.query, _data=df)  # type: ignore

        completed_results.append(final_result)

        return (df, completed_results)  # type: ignore



import logging
import threading
from contextlib import nullcontext
from dataclasses import dataclass
from typing import Dict, Iterable, Optional

import duckdb

# from pandas import DataFrame
from pandas import ArrowDtype
from pyarrow import Table

from ..datamodel.database import IqlDatabase, IqlDatabaseConnector, IqlResult

logger = logging.getLogger(__name__)


def rename_dupe_arrow_columns(pat: Table) -> Table:
    # PyArrow can't export to Pandas if there are duplicate columns
    cols = pat.column_names
    unique_cols = set(cols)

    if len(unique_cols) == len(cols):
        # everything is unique
        return pat
    else:
        logger.warning("Has dupes, %s", cols)

        newcols = []
        for col in cols:
            idx = 1
            newcol = col

            while newcol in newcols:
                newcol = f"{col}_{idx}"
                idx += 1

            newcols.append(newcol)

        pat = pat.rename_columns(newcols)

        return pat


@dataclass
class DuckDbResult(IqlResult):
    _table: Table

    def arrow(self):
        return self._table

    def df(self):
        return self._table.to_pandas(types_mapper=ArrowDtype)

    def native(self):
        return self._table


@dataclass
class _DuckDB(IqlDatabase):
    _connection: object
    _started_thread: int

    def execute_query(
        self,
        query: str,
        context: Optional[Dict[str, object]],  # Duckdb uses the caller frame to auto-populate variables
        completed_results: Optional[Iterable[IqlResult]] = None,
        raw: bool = False,
        parameters: Optional[Iterable] = None,
    ) -> Optional[DuckDbResult]:
        """param: Each of the completed _dfs are registered to the database.
        threaded: pass True to run in a separate connection, otherwise runs in main connection
        """
        if context is not None:
            # TODO: Don't do this... this was because duckdb stopped using the full context.
            # Alternative is to use full chain, rather than just caller frame
            locals().update(context)

        if len(query.strip()) == 0:
            return None

        threaded = threading.get_ident() != self._started_thread

        # create & close the connection if we're threading, otherwise create one
        with self.get_connection() if threaded else nullcontext(self._connection) as con:
            try:
                # Register each of the dataframes to duckdb, so duckdb can query them
                # Other database might require a "load" or "from_pandas()" step to load these
                # to temporary tables.
                if completed_results is not None:
                    for result in completed_results:
                        data = result.native()
                        if data is not None:  # None for paths
                            con.register(result.name, data)  # type: ignore

                if raw:  # for DBT
                    rel = con.execute(query, parameters=parameters)  # type: ignore
                    return DuckDbResult(name=query, query=query, _table=rel)
                else:
                    d = con.sql(query, params=parameters)  # type: ignore

                    if d is not None:
                        table = d.arrow()
                        table = rename_dupe_arrow_columns(table)
                        return DuckDbResult(name=query, query=query, _table=table)

                    else:
                        return None
            except Exception as e:
                e.add_note(f"Query:\n {query}")
                raise

    def get_connection(self):
        return self._connection.cursor()  # type: ignore

    def close_db(self):
        self._connection.close()  # type: ignore
        self._connection = None


class _DuckDbConnector(IqlDatabaseConnector):
    def create_database(self, file=":memory:") -> _DuckDB:
        con = duckdb.connect(database=file)
        # con.execute("SET enable_progress_bar=false;")
        return _DuckDB(_connection=con, _started_thread=threading.get_ident())

    def create_database_from_con(self, con: object) -> _DuckDB:
        return _DuckDB(_connection=con, _started_thread=threading.get_ident())


def get_connector() -> IqlDatabaseConnector:
    return _DuckDbConnector()



import logging
from abc import abstractmethod
from typing import Dict, Iterable, Optional

from .result import IqlResult

logger = logging.getLogger(__name__)


class IqlDatabase:
    @abstractmethod
    def execute_query(
        self,
        query: str,
        context: Optional[Dict[str, object]],
        completed_results: Iterable[IqlResult],
        raw: bool = False,
        parameters: Optional[Iterable] = None,
    ) -> Optional[IqlResult]:
        pass

    @abstractmethod
    def get_connection(self) -> object:
        pass

    @abstractmethod
    def close_db(self):
        pass


class IqlDatabaseConnector:
    @abstractmethod
    def create_database(self) -> IqlDatabase:
        pass

    @abstractmethod
    def create_database_from_con(self, con: object) -> IqlDatabase:
        pass



from typing import Dict

# Add extensions via register_extension()
# Extensions are loaded on first access, to avoid requiring
# unused dependencies

_KNOWN_EXTENSIONS: Dict[str, str] = {
    "bql": "iql.extensions.bql_ext.bql_extension",
    "pandas": "iql.extensions.pandas_extension",
    "cache": "iql.extensions.cache_extension",
    "blpapi": "iql.extensions.blpapi_ext.blp_extension",
    "azureblob": "iql.extensions.azure_blob_extension",
    "nestiql": "iql.extensions.nest_extension",
    "openai": "iql.extensions.openai_extension",
}

_LOADED_EXTENSIONS = []



import logging
import re
from dataclasses import dataclass, field
from typing import Dict, Iterable, Optional, Union

from iql.extensions.bql_ext.bql_wrapper import (
    BaseBqlQuery,
    construct_bql_query,
    security_to_finalstr,
)

logger = logging.getLogger(__name__)

SECURITY_KEYWORD = "$SECURITY"


@dataclass
class RawBqlQuery(BaseBqlQuery):
    """A full BQL query. Ideally, use $SECURITY instead of a single equity name"""

    fieldpattern = re.compile(r"(?s).*get\s*\((.*?)\)\s*for.*")

    bql_query_string: str
    security: Optional[str] = None
    params: Dict = field(default_factory=dict)  # parameters. Convention is $FIELD: value

    def get_fields(self) -> Iterable[str]:
        """Extracts individual fields from the BQL get statement.
        TODO: Handle unbalanced escaped/quoted parens"""

        query = self.to_bql_query()
        match = self.fieldpattern.fullmatch(query)

        if match is None:
            raise ValueError(f"Couldn't extract GET from: {query}")

        else:
            getclause = match.group(1)
            fields = []
            start = 0
            depth = 1
            for i, c in enumerate(getclause):
                if c == "," and depth == 1:
                    fields.append(getclause[start:i].strip())
                    start = i + 1
                if c == "(":
                    depth += 1
                if c == ")":
                    depth -= 1

            fields.append(getclause[start:].strip())

        return fields

    def to_bql_query(self) -> str:
        if self.security is not None:
            new_str = self.bql_query_string.replace(SECURITY_KEYWORD, self.security)
        else:
            new_str = self.bql_query_string

        for param, value in self.params.items():
            new_str = new_str.replace(param, value)

        logger.debug("Raw query to string: %.50s", new_str)

        if "preferences" not in new_str:
            logger.debug("Adding default preferences")
            new_str += " \n preferences(addcols=all)"
        return new_str


@dataclass
class BqlQuery(BaseBqlQuery):
    name: str
    fields: Iterable[str]

    security: Union[str, Iterable]
    let_str: Optional[str]
    with_params: Optional[str]

    for_str: str = SECURITY_KEYWORD
    params: Dict = field(default_factory=dict)  # parameters. Convention is $FIELD: value

    def get_fields(self) -> Iterable[str]:
        return self.fields

    def to_bql_query(self) -> str:
        for_str_mod = security_to_finalstr(self.for_str)

        if SECURITY_KEYWORD not in self.params and self.security is not None:
            self.params[SECURITY_KEYWORD] = self.security

        sec = self.params.get(SECURITY_KEYWORD)
        if self.let_str is not None and sec is not None and SECURITY_KEYWORD in self.let_str:
            let_str_mod = self.let_str.replace(SECURITY_KEYWORD, sec)
        else:
            let_str_mod = self.let_str
        query_str = construct_bql_query(self.fields, for_str_mod, self.with_params, let_str_mod)

        logger.debug("After construction but before replacement: %s", query_str)
        for param, value in self.params.items():
            if param == SECURITY_KEYWORD:
                # Don't sub if:
                # is a list
                # is a function (bracket or paren)
                # is already quoted
                if not (isinstance(value, str) and ("'" in self.for_str or "(" in self.for_str or "[" in self.for_str)):
                    value = security_to_finalstr(self.params[SECURITY_KEYWORD])

            query_str = query_str.replace(param, value)

        logger.debug("BQL Query to String: %s", query_str)
        return query_str



import logging
from dataclasses import dataclass
from typing import Optional

import pyarrow as pa

from ...datamodel.extension import IqlExtension
from ...datamodel.subquery import SubQuery
from ...ql import register_extension
from . import blp_api_lowlevel as bal

logger = logging.getLogger(__name__)


@dataclass
class BlpApiListFieldsExtension(IqlExtension):
    param_replace_text = False

    def executeimpl(self, sq: SubQuery) -> Optional[pa.Table]:
        if len(sq.options) == 0:
            raise ValueError('Must select a field, or "All" for all fields')

        search = sq.options.get("search", False)

        if sq.get_query() == "All":
            logger.debug("Listing all fields")
            fi = bal.get_field_info()
        else:
            fi = bal.get_field_info(sq.get_query(), search=bool(search))

        t = pa.Table.from_pylist(fi)
        return t


@dataclass
class BlpApiEntitlementsExtension(IqlExtension):
    param_replace_text = False

    def executeimpl(self, sq: SubQuery) -> Optional[pa.Table]:
        if sq.options is None or len(sq.options) == 0 or sq.get_query() is None:
            raise ValueError("UUID required")

        e = bal.get_user_entitlements(sq.get_query())

        t = pa.Table.from_pylist(e)
        return t


@dataclass
class BlpApiRefDataExtension(IqlExtension):
    param_replace_text = False

    def executeimpl(self, sq: SubQuery) -> Optional[pa.Table]:
        fields = sq.options.get("fields", None)

        if fields is None:
            raise ValueError("Must pass a list of fields to retrieve")

        securities = sq.options.get("securities", None)
        if securities is None:
            raise ValueError("Must pass a list of securities")

        if isinstance(fields, str):
            fields = (fields,)

        params: dict = sq.options.get("params", {})  # type: ignore
        for k, v in sq.options.items():
            # Add any explicit parameters, so everything doesn't need to be packed into a dict
            if v is not None and k not in (
                "fields",
                "securities",
                "historical",
                "params",
                "paramquerybatch",
                "paramquery",
                "pivot",
            ):
                params[k] = v

        historical = sq.options.get("historical", False)

        if "startDate" in params or "endDate" in params:
            logger.debug("Historical enabled because startDate or endDate in params")
            historical = True

        rd = bal.get_refdata(
            fields=fields,
            securities=securities,
            parameters=params,
            historical=historical,  # type: ignore
        )

        if len(rd) == 0:
            raise ValueError(f"Empty response for query {sq}")
        else:
            t = pa.Table.from_pylist(rd)
            return t


def register(keyword: str):
    extension = BlpApiListFieldsExtension(keyword=keyword, subword="fields")
    register_extension(extension)

    extension = BlpApiEntitlementsExtension(keyword=keyword, subword="userentitlements")
    register_extension(extension)

    extension = BlpApiRefDataExtension(keyword=keyword)
    register_extension(extension)



# Copyright (C) 2025, IQMO Corporation [support@iqmo.com]
# All Rights Reserved

import logging
import os
from dataclasses import dataclass
import struct 
import pyodbc
from pandas import DataFrame, read_sql

from ... import IqlExtension, SubQuery, register_extension
from ...datamodel import cache

logger = logging.getLogger(__name__)
from sqlalchemy import create_engine
import sqlalchemy as sa
from sqlalchemy.engine import URL

# https://github.com/pymssql/pymssql
# https://pymssql.readthedocs.io/en/stable/

_CREDENTIAL = None
replace_pyformat_parameters = True

# The connection cache is either connections **or** options to create a connection
_CONNECTION_CACHE: dict[str, sa.Engine | dict] = {}
CACHE_INVALIDATION_QUERY = None


def _get_connection_from_options(connection_string: str, **kwargs) -> sa.Engine:
        
    connection_url = URL.create("mssql+pyodbc", query={"odbc_connect": connection_string})

    sqltoken = _CREDENTIAL.get_token("https://database.windows.net/.default").token

    # https://learn.microsoft.com/en-us/azure/azure-sql/database/azure-sql-python-quickstart?view=azuresql&tabs=windows%2Csql-inter
    raw = sqltoken.encode("utf-16-le")
    token_bytes = struct.pack(f"<I{len(raw)}s", len(raw), raw)
    tokenshort = {1256: token_bytes}
    #result = iql.execute(f"select * from mssql.pyodbc_connect(connection_string = '{CONN_URL}', attrs_before={tokenshort})", con=conn)
    kwargs["attrs_before"]=tokenshort

    engine = create_engine(connection_url, connect_args=kwargs)

    return engine

def _get_connection(connection_name: str) -> pyodbc.Connection:
    existing_conn: sa.Engine | dict = _CONNECTION_CACHE[connection_name]
    if isinstance(existing_conn, dict):
        return _get_connection_from_options(**existing_conn)
    else:
        return existing_conn

def process_result(result: pyodbc.Cursor):
    if result.description:                
        columns = [col[0] if col[0] else f"col_{i}" for i, col in enumerate(result.description)]
        result_df = DataFrame.from_records(result.fetchall(), columns=columns)
    else:
        row_count = result.rowcount if result.rowcount!=-1 else None
        result_df = DataFrame({"result": ["query complete"], "rowcount": [row_count]})
    
    return result_df
        

def _execute_query(engine: sa.Engine, query: str, parameters: dict | None = None) -> list[DataFrame]:
    logger.debug("Executing query in sqlalchemy: %s", query)
    
    
    results = []
    with engine.raw_connection().cursor() as conn:

        if parameters is not None:
            if replace_pyformat_parameters:
                for k,v in parameters.items():
                    query=query.replace(f"%({k})s", f"'{v}'")

                result = conn.execute(query)

            else:
                result = conn.execute(query, parameters)
        else:
            result = conn.execute(query)

        while result:
            result_df = process_result(result)
            results.append(result_df)
            
            if not result.nextset():
                break

        return results
    

@dataclass
class SqlServerExtensionPyOdbcConnect(IqlExtension):
    def executeimpl(self, sq: SubQuery) -> DataFrame:
        connection_name: str = sq.options.get("name", "default")  # type: ignore
        connection_string: str = sq.options["connection_string"]  # type: ignore

        eager = sq.options.get("eager", True)
        conn_options = {
            k: v for k, v in sq.options.items() if k != "name" and k != "eager" and k != "connection_string"
        }

        if eager:
            _CONNECTION_CACHE[connection_name] = _get_connection_from_options(
                connection_string=connection_string, **conn_options
            )
        else:
            _CONNECTION_CACHE[connection_name] = conn_options

        return DataFrame({"success": [True], "message": ["Connection Successful"]})


@dataclass
class SqlServerExtensionPyOdbc(IqlExtension):
    @cache.iql_cache
    def executeimpl(self, sq: SubQuery) -> DataFrame:
        connection_name: str = sq.options.get("name", "default")  # type: ignore

        conn: pyodbc.Connection = _get_connection(connection_name)

        query: str = sq.get_query()  # type: ignore

        parameters: dict = sq.options.get("PARAMETERS", None)  # type: ignore

        try:
            results = _execute_query(conn, query, parameters=parameters)
        except Exception as e:
            e.add_note(f"{query=}")
            raise
        return results[-1]
        

def register(keyword: str):
    if CACHE_INVALIDATION_QUERY is not None:
        logger.info("Using CACHE_INVALIDATION_QUERY")
        ext_cache = cache.QueryInvalidationCache(CACHE_INVALIDATION_QUERY)
    else:
        ext_cache = None

    extension = SqlServerExtensionPyOdbc(keyword=keyword, cache=ext_cache)
    register_extension(extension)

    extension = SqlServerExtensionPyOdbcConnect(keyword=keyword, subword="pyodbc_connect", cache=None)
    register_extension(extension)

    extension = SqlServerExtensionPyOdbc(keyword=keyword, subword="pyodbc", cache=ext_cache)
    register_extension(extension)

    extension = SqlServerExtensionPyOdbc(keyword=keyword, subword="nocache", cache=None)
    register_extension(extension)


import logging
from dataclasses import dataclass

logger = logging.getLogger(__name__)


@dataclass
class IqlResult:
    name: str
    query: str

    def arrow(self):
        raise NotImplementedError("Not Implemented")

    def df(self):
        raise NotImplementedError("Not Implemented")

    def native(self):
        """Returns whatever the internal representation"""
        raise NotImplementedError("Not Implemented")


@dataclass
class IqlResultData(IqlResult):
    name: str
    query: str
    _data: object

    def native(self):
        """Returns whatever the internal representation"""
        return self._data

    def arrow(self):
        raise NotImplementedError("Not Implemented")

    def df(self):
        raise NotImplementedError("Not Implemented")



import asyncio
import hashlib
import logging
from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Callable, Dict, Iterable, Optional, Tuple, Union

from pandas import DataFrame

from .. import ql
from ..datamodel.cache import NoopSqCache, SqCache
from .querycontainer import IqlQueryContainer
from .result import IqlResultData
from .subquery import SubQuery, SubQueryGroup, get_subqueries_flat

logger = logging.getLogger(__name__)


@dataclass
class IqlExtension:
    keyword: str
    subword: str = field(default=None, init=True)  # type: ignore
    is_async: bool = False  # Determines whether executeimpl is implemented asynchronously
    cache: SqCache | None = field(default_factory=NoopSqCache, init=True)  # default is a NoopCache

    # Extensions must be parallelizable, or must
    # implement their own execute_batch.

    # Determines whether the local cache settings should be used
    # vs  CACHE_PERIOD and
    # USE_FILE_CACHE
    temp_file_directory: Optional[str] = field(default=None, init=False)
    param_replace_text: bool = field(
        default=True, init=False
    )  # determines whether paramquery/paramquerybatch will replace the query, or add parameters

    def get_output_dir(self) -> str:
        tempdir = self.temp_file_directory if self.temp_file_directory is not None else "./"
        return tempdir

    def use_path_replacement(self) -> bool:
        """Some extensions just return a filestring to use instead of the SubQuery, such as the
        S3 Extension. If execute() returns None, then use_path_replacement must be used.
        """
        return False

    def get_path_replacement(self, subquery: SubQuery, quote: bool = True) -> Optional[str]:
        """
        Returns a path to a file string that is readable by the underlying database: parquet or csv, most commonly.

        Quoting is done here, with the expectation that some path replacements might actually be function calls.
        """
        tempdir = self.get_output_dir()

        if subquery is None:  # subquery group
            prehash = "\n".join(sq.subquery for sq in subquery.subqueries)
            sq_hash = hashlib.sha256(prehash.encode()).hexdigest()
        else:
            if subquery.subquery is None:
                sq_str = ""
            else:
                sq_str = subquery.subquery
            sq_hash = hashlib.sha256(sq_str.encode()).hexdigest()

        filepath = f"{tempdir}/{subquery.name}_{sq_hash}.parquet"

        if quote:
            outpath = f"'{filepath}'"
        else:
            outpath = filepath

        logger.debug("Converted %s to %s", subquery, outpath)
        return outpath

    @abstractmethod
    def executeimpl(self, sq: SubQuery) -> Optional[DataFrame]:
        raise NotImplementedError("Not Implemented")

    async def executeasync(self, sq: SubQuery) -> Optional[DataFrame]:
        if sq.dataframe is not None:
            return sq.dataframe

        df = await self.executeimpl(sq=sq)

        sq.dataframe = df
        return df

    def execute(self, sq: SubQuery) -> Optional[DataFrame]:
        # usage: select * from (verityapi(functionname, targetname)) as verityquery
        # An empty response means no response was needed
        # Internal failure must raise an exception

        if sq.dataframe is not None:
            return sq.dataframe

        logger.debug("Executing query %s", sq.subquery)

        result_df = self.executeimpl(sq)

        if self.use_path_replacement():
            return None
        else:
            if result_df is None:
                raise ValueError("Unexpected empty result")

            pivot_options = sq.options.get("pivot")
            melt_options: dict = sq.options.get("melt")  # type: ignore
            result_df = self.apply_pandas_operations_prepivot(result_df, sq.options)

            if pivot_options is not None:
                result_df = self.pivot(result_df, pivot_options)  # type: ignore
            if melt_options is not None:
                result_df = result_df.melt(**melt_options)

            result_df = self.apply_pandas_operations_postpivot(result_df, sq.options)

            sq.dataframe = result_df

            return result_df

    def execute_batch(self, queries: Iterable[SubQuery]) -> Iterable[IqlResultData]:
        """This is a default implementation that executes each of the queries serially.

        Extensions may override this for cases that could be executed concurrently, such as BQL.

        """
        if self.is_async:
            # Execute as a batch
            # If run in Jupyter, needs nest_async to allow nested async.
            subqueries_flat = get_subqueries_flat(queries)

            logger.debug("Executing batch of %s queries", len(list(queries)))

            # Avoid any issue with nesting asyncio
            logger.info("Running in separate pool")

            async def _run_tasks(subqueries_flat):
                return await asyncio.gather(*(self.executeasync(sq) for sq in subqueries_flat))

            results = asyncio.run(_run_tasks(subqueries_flat))

            for sq, df in zip(subqueries_flat, results, strict=False):
                sq.dataframe = df
        else:
            # Execute serially
            # context = get_context('spawn')

            subqueries_flat = get_subqueries_flat(queries)

            # Executes the ungrouped queries
            for query in subqueries_flat:
                try:
                    self.execute(query)  # type: ignore
                except Exception as e:
                    logger.exception("Error in %s", query)
                    raise e

        # Merge the results
        completed_results = []

        for query in queries:
            # For single subqueries, this is a noop
            # For subquery groups, this merges the results of the children
            query.merge()

            if self.use_path_replacement():
                continue
            else:
                ir = IqlResultData(name=query.name, query=query.subquery, _data=query.dataframe)
                completed_results.append(ir)

        return completed_results

    def create_subquery(self, query: str, name: str, iqc: IqlQueryContainer) -> SubQuery:
        logger.debug("Creating subquery")
        sq = SubQuery(extension=self, subquery=query, name=name, iqc=iqc)
        return sq

    def update_replacement_values(self, replacement_values, parameter_name, parameter_values):
        if len(replacement_values) == 0:
            # If there was no paramquery or paramquery_batch, then this is just a paramlist
            for value in parameter_values:
                replacement_values.append({parameter_name: value})
        else:
            # We have some replacements already
            new_replacement_values = []
            for replacement_list in replacement_values:
                for value in parameter_values:
                    rl_copy = replacement_list.copy()

                    rl_copy[parameter_name] = value
                    new_replacement_values.append(rl_copy)
            replacement_values = new_replacement_values

        return replacement_values

    def create_subqueries(  # noqa: C901
        self,
        query: str,
        name: str,
        iqc: IqlQueryContainer,
        create_function: Optional[Callable] = None,
    ) -> Iterable[SubQuery]:
        """
        Implementation note:
        Current implementation requires the extension to "merge" the results, if there are multiple subqueries
        Previous implementation used a database union to let the db merge the results. This is still do-able, but

        Supports:
        - paramquery(parameter_name, query): Allows one or multiple parameter names to be replaced
        - paramquerybatch(parameter_name, query, max_batch_size): Only a single parameter_name is supported, default is
        unlimited max_batch_size -1
        - paramlist(parameter_name, parameter_values): Substitutes a fixed list of values into parameter_name
        """

        logger.info("Creating a subquery %.50s for %.50s", name, query)

        if create_function is None:
            create_function = self.create_subquery

        sq = create_function(query=query, name=name, iqc=iqc)

        # If a paramlist is passed, create one subquery for each value
        replacement_values: Iterable[Dict[str, object]] = []

        if "paramquerybatch" in sq.options:
            # paramquerybatch(targetfield, sourcequery, max_batch_size).
            # max_batch_size defaults to unlimited
            paramquery: Union[Tuple[str, str], Tuple[str, str, int]] = sq.options.get("paramquerybatch")  # type: ignore

            parameter_name = paramquery[0]
            param_query = paramquery[1]

            if len(paramquery) == 3:
                max_batch_size = paramquery[2]
            else:
                max_batch_size = -1

            param_query_batch_df = ql.execute(param_query, con=iqc.db._connection, substitutions=iqc.substitutions)  # type: ignore

            parameter_query_values: Iterable = param_query_batch_df[param_query_batch_df.columns[0]]  # type: ignore
            parameter_query_values = sorted(parameter_query_values)
            parameter_values = []
            if max_batch_size <= 0:
                parameter_values.append(parameter_query_values)  # type: ignore
            else:
                for i in range(0, len(parameter_query_values), max_batch_size):  # type: ignore
                    parameter_values.append(
                        parameter_query_values[i : i + max_batch_size]  # type: ignore
                    )
            parameter_values_list = [list(batch) for batch in parameter_values]  # type: ignore
            for param_value in parameter_values_list:
                replacement_values.append({parameter_name: param_value})

            logger.debug(
                "Max batch size is %s, Creating batches %s",
                max_batch_size,
                len(parameter_values),
            )
            logger.debug("Values: %s", parameter_values)

        if "paramquery" in sq.options:
            # Accepts:
            # parameter_name: Union[str, Iterable[str]]
            # query: str
            paramquery: Tuple[str, str] = sq.options.get("paramquery")  # type: ignore
            if len(paramquery) != 2:
                raise ValueError(f"A paramquery accepts exactly two parameters, yet {len(paramquery)} were passed")

            parameter_name = paramquery[0]
            param_query = paramquery[1]

            param_query_df: DataFrame = ql.executedf(
                param_query,
                con=iqc.db.get_connection(),
                substitutions=iqc.substitutions,
            )

            if isinstance(parameter_name, str):
                parameter_values = param_query_df[param_query_df.columns[0]]  # type: ignore
                replacement_values = self.update_replacement_values(
                    replacement_values, parameter_name, parameter_values
                )

            else:  # multiple parameters, so take the results sequentially from the df
                parameter_values = [
                    *param_query_df.itertuples(index=False, name=None)  # type: ignore
                ]
                for t in parameter_values:
                    row = {}
                    for i, name in enumerate(parameter_name):
                        row[name] = t[i]
                    replacement_values.append(row)

                logger.debug("replacement_values = %s", replacement_values)

        if "paramlist" in sq.options:
            logger.info("Doing paramlist")
            paramlist: Tuple[str, Iterable[str]] = sq.options.get("paramlist")  # type: ignore

            if len(paramlist) != 2:
                raise ValueError(f"A paramlist accepts exactly two parameters, yet {len(paramlist)} were passed")

            parameter_name = paramlist[0]
            parameter_values: Iterable[str] = paramlist[1]

            if not isinstance(parameter_values, Iterable):
                raise ValueError("A paramlist parameter_values must be an Iterable")

            logger.debug("Found paramlist: %s => %s", parameter_name, parameter_values)

            replacement_values = self.update_replacement_values(replacement_values, parameter_name, parameter_values)

            if parameter_values is None or len(parameter_values) == 0:
                raise ValueError("Empty values passed to paramlist passed")

        if len(replacement_values) > 0:
            sqs: Iterable[SubQuery] = []
            count = 1

            # Some extensions, like bql, use string replacement.
            # Others, like blpapi, use named parameters
            replace_text = sq.extension.param_replace_text

            for replacements in replacement_values:
                v_query = query

                # strip off any parameters
                new_options = {}
                for parameter_name, parameter_value in replacements.items():
                    if replace_text:
                        if isinstance(parameter_value, list):
                            parameter_value = ",".join([f"'{v}'" for v in parameter_value])
                        if not parameter_name.startswith("$"):
                            parameter_name = f"${parameter_name}"
                        v_query = v_query.replace(parameter_name, str(parameter_value))
                    else:
                        new_options[parameter_name] = parameter_value
                newsq = create_function(  # type: ignore
                    query=v_query, name=f"{name}_{count}", iqc=iqc
                )

                newsq.options.update(new_options)
                sqs.append(newsq)
                count += 1

            sqname = name.replace("$", "")  # safeguard if user inserts the optional $ prefix

            if len(sqs) == 1:
                return sqs  # Don't need a subquery group if it's just one value
            else:
                sq_group = SubQueryGroup(
                    subqueries=sqs,
                    extension=self,
                    subquery=sqname,  # subqueries are ignored for subquerygroups
                    name=f"{sqname}_group",
                    iqc=iqc,
                )

                logger.debug("sq_group=%s", sq_group)
                return [sq_group]
        else:
            return [sq]

    def fix_col_ref(self, opt: str, columns: Iterable[str]):
        if opt in columns:
            return opt
        opt_l = opt.lower().strip()
        opt_ci = next((c for c in columns if c.lower() == opt_l), None)

        if opt_ci is None:
            raise ValueError(f"{opt} not in columns: {columns}")

        return opt_ci

    def apply_pandas_operations_postpivot(self, working_df, options: Dict[str, object]) -> DataFrame:
        fillna_opt: str = options.get("fillna")  # type: ignore
        if fillna_opt is not None:
            working_df = working_df.fillna(fillna_opt)

        dropna_opt = options.get("dropna")
        if isinstance(dropna_opt, bool) and dropna_opt is True:
            working_df = working_df.dropna()
        elif isinstance(dropna_opt, str):
            working_df = working_df.dropna(subset=[dropna_opt])
        elif isinstance(dropna_opt, list):
            working_df = working_df.dropna(subset=dropna_opt)

        if "ewma" in options:
            ewma_opts: Iterable[Tuple[str, str, dict[str, object]]] = options.get("ewma")  # type: ignore

            for ewma_opt in ewma_opts:
                src, dest, options = ewma_opt
                working_df[dest] = working_df[src].ewm(**options).mean()

        return working_df

    def apply_pandas_operations_prepivot(self, working_df, options: Dict[str, object]) -> DataFrame:
        addcols_opt: Iterable[str] = options.get("addcols_pre")  # type: ignore
        if addcols_opt is not None:
            if isinstance(addcols_opt, str):
                addcols_opt = [addcols_opt]
            for col in addcols_opt:
                if col not in working_df:
                    logger.debug("Adding additional columns (before pivoting): %s", col)
                    working_df[col] = None

        # only drops from the "value" column
        fillna_opt: str = options.get("fillna_pre")  # type: ignore
        if fillna_opt is not None:
            logger.debug("Filling NaNs with in value column with %s", fillna_opt)
            working_df["value"] = working_df["value"].fillna(fillna_opt)

        dropna_opt = options.get("dropna_pre")
        logger.debug("Dropping NA from column %s", dropna_opt)
        if isinstance(dropna_opt, bool) and dropna_opt is True:
            working_df = working_df.dropna()
        elif isinstance(dropna_opt, str):
            working_df = working_df.dropna(subset=[dropna_opt])
        elif isinstance(dropna_opt, list):
            working_df = working_df.dropna(subset=dropna_opt)

        return working_df

    def pivot(self, working_df: DataFrame, pivot_option: Iterable[str]) -> DataFrame:  # noqa: C901
        if pivot_option is None:
            return working_df

        pivot_option = list(pivot_option)

        logger.debug("Pivoting by %s", pivot_option)
        if len(pivot_option) != 2 and len(pivot_option) != 3:
            raise ValueError(f"Unexpected size for pivot options {pivot_option}")
        else:
            index = pivot_option[0]
            column = pivot_option[1]
            value = pivot_option[2] if len(pivot_option) == 3 else "value"

            if index == "auto":
                if isinstance(column, str):
                    used_cols = [column.lower()]
                else:
                    used_cols = list(column)

                if isinstance(value, str):
                    used_cols.append(value.lower())
                else:
                    used_cols = used_cols + value

                # Use all columns except used_cols
                index = [col for col in working_df.columns if col.lower() not in used_cols]
        cols: Iterable[str] = list(working_df.columns)  # type: ignore

        if isinstance(index, list):
            index = [self.fix_col_ref(i, cols) for i in index]
            if len(index) == 1:
                index = index[0]
        else:
            index = self.fix_col_ref(index, cols)

        if isinstance(column, list):
            column = [self.fix_col_ref(i, cols) for i in column]
            if len(column) == 1:
                column = column[0]
        else:
            column = self.fix_col_ref(column, cols)

        if value == "bqlvalue":
            # Special case
            pass

        elif isinstance(value, list):
            value = [self.fix_col_ref(i, cols) for i in value]
            if len(value) == 1:
                value = value[0]
        else:
            value = self.fix_col_ref(value, cols)

        # Disabling: Unsure how this is reached
        # if isinstance(column, list):
        #    for col in column:
        #        if pd.api.types.is_datetime64_any_dtype(working_df[col]):
        #            working_df[col] = working_df[col].dt.strftime("%Y-%m-%d")

        logger.debug("Pivot index %s columns %s values %s", index, column, value)

        if value == "bqlvalue":
            value = [c for c in working_df.columns if c.startswith("value_")]
            # adding value would result in multiples
            # #value.append("value")
            skip_flatten_use_second_row_only = True
            aggfunc = "first"  # use the first column in the value list
            logger.debug("Using first value from %s for pivot", value)
            dropna = True
            reset_first = False
        else:
            aggfunc = "last"
            skip_flatten_use_second_row_only = False
            dropna = False
            reset_first = True

        # figure out what columns *should* show up, in case they get dropped if all NA
        # The problem is: pivot_table will add dummy rows for every index permutation,
        # so dropna is needed... but this may drop some valid columns
        if isinstance(column, str):
            all_columns = list(working_df[column].unique())
        else:
            all_columns = list()

        working_df = working_df.pivot_table(index=index, columns=column, values=value, aggfunc=aggfunc, dropna=dropna)

        if reset_first:
            working_df = working_df.reset_index()

        if skip_flatten_use_second_row_only:
            working_df.columns = [col[1] for col in working_df.columns]
            working_df = working_df.reset_index()

        elif isinstance(value, list) and len(value) > 1:
            working_df.columns = [
                "_".join(reversed(str(col) if isinstance(col, int) else col)) for col in working_df.columns
            ]
        elif isinstance(column, list) and len(column) > 1:
            # Flatten multicolumn indices
            working_df.columns = ["_".join(str(col) if isinstance(col, int) else col) for col in working_df.columns]

        # Reinsert any dropped all NA columns
        for col in working_df.columns:
            if col not in all_columns:
                all_columns.append(col)

        working_df = working_df.reindex(columns=all_columns, fill_value=None)

        # Clean columns
        renames = {}
        for col in working_df.columns:
            # assert isinstance(col, str)
            colname = col
            newcol = str(colname).replace("#", "").replace("(", "").replace(")", "").replace(" ", "_")
            newcol = newcol.strip("_")
            if col != newcol:
                # columns can't have #, ( or ) symbols
                renames[col] = newcol

        if len(renames) > 0:
            working_df = working_df.rename(columns=renames)

        return working_df



import ast
import logging
from typing import Dict, Iterable, Union

logger = logging.getLogger(__name__)

lastnode = None


def convert(node) -> Union[Iterable, object]:
    global lastnode
    lastnode = node
    if hasattr(node, "elts"):
        v = []
        for e in node.elts:
            val = convert(e)
            v.append(val)
        return v
    elif isinstance(node, ast.Constant):
        return node.value
    elif isinstance(node, ast.Dict):
        return {convert(key): convert(value) for key, value in zip(node.keys, node.values)}
    else:
        return node.id


def options_to_list(options: str) -> Dict[str, object]:
    p = ast.parse(options)

    results = {}

    for e in p.body:
        # This code is never reached now?
        # if isinstance(e.value, ast.Name):  # type: ignore
        #    results[e.value] = None  # type: ignore

        #    continue
        # else:
        for arg in e.value.args:  # type: ignore
            results[arg.value] = None

        for k in e.value.keywords:  # type: ignore
            r = convert(k.value)

            results[k.arg] = r

            continue

    return results



import logging
from dataclasses import dataclass, field
from typing import Iterable, Optional

from pandas import DataFrame

from ...datamodel import cache
from ...datamodel.extension import IqlExtension
from ...datamodel.querycontainer import IqlQueryContainer
from ...datamodel.result import IqlResultData
from ...datamodel.subquery import SubQuery, get_subqueries_flat
from ...ql import execute, register_extension
from .bql_datamodel import RawBqlQuery
from .bql_wrapper import BaseBqlQuery, execute_bql_str_list_async_q

logger = logging.getLogger(__name__)

_KEYWORD = "bql"

# _bql_start_pattern = r"(?si)\(\s*((get)|(let))\s*\("
# _bql_pat = re.compile(_bql_start_pattern)

# _QUERY_FOR_PATTERN = re.compile(
#     r"(?s)(.*for\s*\((.*?)\)\s*)((with.*?)?)\s*((preferences.*)?)"
# )


@dataclass
class _IqlBqlQuery(SubQuery):
    # Extended BQL language to add "iqmo" options
    # such as "splitid" and pivoting
    # Syntax; pivot(id, name)
    # Syntax 2: pivot([id:col2], name)
    bqlquery: BaseBqlQuery = field(init=False)
    force_file: bool = (
        True  # for large queries, forces the subquery to write to a parquet file instead of storing in a df.
    )

    # detects everything before "(.*) as #..."
    def execute(self) -> bool:
        if self.dataframe is not None:
            return True

        success = self.execute_internal()

        if not success:
            logger.warning(self.bqlquery.exception_msg)

        if success:
            # If it's a output file, then it'd already be written
            if not self.extension.use_path_replacement():
                df = self.bqlquery.to_df()

                df = self.extension.apply_pandas_operations_prepivot(df, self.options)

                df = self.extension.pivot(df, self.options.get("pivot"))  # noqa: PD010 # type: ignore
                df = self.extension.apply_pandas_operations_postpivot(df, self.options)

                self.dataframe = df

                if self.extension.cache is not None:
                    self.extension.cache.save(self, df, cost=None)

            return success
        else:
            allow_failure_opt = self.options["allow_failure"]

            if isinstance(allow_failure_opt, str):
                cols: Iterable[str] = [allow_failure_opt]
            else:
                cols: Iterable[str] = allow_failure_opt  # type: ignore
            logger.debug(
                "Query failed, but allow_failure is enabled. Creating an empty dataframe with cols = %s",
                cols,
            )

            self.dataframe = DataFrame(columns=cols)  # type: ignore

            return True

    def execute_internal(self) -> bool:
        if self.bqlquery.execution_status == BaseBqlQuery.STATUS_FAILURE:
            return False

        if self.bqlquery.execution_status == BaseBqlQuery.STATUS_COMPLETED:
            return True

        working_df = None

        # If it already ran, don't run again
        if self.bqlquery.execution_status != BaseBqlQuery.STATUS_COMPLETED:
            self.bqlquery.execute()

        # query has passed
        if working_df is None:
            working_df = self.bqlquery.to_df()

        if working_df is None:
            raise ValueError("Unable to convert query to DataFrame")

        # validate column names: disabled for now, still a work in progress
        # working_df = self.validate_fix_column_names(working_df)
        self.dataframe = working_df
        return True


@dataclass
class BqlExtension(IqlExtension):
    def executeimpl(self, sq: SubQuery) -> Optional[DataFrame]:
        raise NotImplementedError("Not Implemented")

    def create_subquery(self, query: str, name: str, iqc: IqlQueryContainer) -> SubQuery:
        iq = _IqlBqlQuery(subquery=query, name=name, extension=self, iqc=iqc)

        bqlstr = iq.get_query()

        q = RawBqlQuery(bqlstr)  # type: ignore
        iq.bqlquery = q
        return iq

    def create_subqueries(
        self,
        query: str,
        name: str,
        iqc: IqlQueryContainer,
        create_function: object = None,
    ) -> Iterable[SubQuery]:
        return super().create_subqueries(query=query, name=name, iqc=iqc, create_function=self.create_subquery)

    def execute_batch(
        self,
        queries: Iterable[_IqlBqlQuery],
    ) -> Iterable[IqlResultData]:
        # Caching operates at two levels here:
        # The SubQuery result, and within the bql_wrapper
        # To force a refresh, we'll clear both the subquery and bql raw query
        # cache

        subqueries_flat: Iterable[_IqlBqlQuery] = get_subqueries_flat(queries)  # type: ignore

        for sq in subqueries_flat:
            # if cached, used cached
            if sq.dataframe is None and self.cache is not None:
                cached_df = self.cache.get(sq)
                if cached_df is not None:
                    sq.dataframe = self.cache.get(sq)

        to_run: Iterable[_IqlBqlQuery] = [sq for sq in subqueries_flat if sq.dataframe is None]

        if self.use_path_replacement():
            for q in to_run:
                q.bqlquery._output_filename = self.get_path_replacement(q, quote=False)

        queries_l: Iterable[BaseBqlQuery] = [q.bqlquery for q in to_run]

        logger.debug("Executing %s", len(queries_l))

        if self.use_path_replacement():
            max_concurrent_queries = 16
            execute_bql_str_list_async_q(queries_l, max_queries=max_concurrent_queries, cache=self.cache)
        else:
            execute_bql_str_list_async_q(queries_l, cache=self.cache)

        # Then update the subqueries
        for q in subqueries_flat:
            if q.bqlquery.execution_status == BaseBqlQuery.STATUS_FAILURE:
                if q.options.get("allow_failure") is None:
                    raise ValueError(
                        f"BQL SubQuery failed {q.bqlquery.exception_msg}: {q.bqlquery.to_bql_query()} {q.options}"
                    )
            try:
                q.execute()
            except Exception as e:
                raise ValueError(f"Error on q.bqlquery: {q.bqlquery.to_bql_query()}") from e

        completed_results = []
        # Final step to merge any grouped subqueries.
        for q in queries:
            q.merge()
            df = q.dataframe

            result = IqlResultData(name=q.name, query=q.subquery, _data=df)

            completed_results.append(result)

        return completed_results


def execute_bql(query: str, pivot: Optional[str] = None) -> DataFrame:
    """Convenience function for testing, or executing single queries"""

    if _KEYWORD in query:
        raise ValueError(f"Pass raw BQL queries only, don't wrap with {_KEYWORD}(...)")

    elif pivot is not None:
        suffix = f", pivot={pivot}"

    else:
        suffix = ""

    query = f'{_KEYWORD}("{query}"{suffix})'
    extension = BqlExtension(_KEYWORD)

    logger.debug("Converted to: %s", query)
    sqs = extension.create_subqueries(query, "Anon", iqc=None)  # type: ignore
    sq = next(sq for sq in sqs)  # there'll only be one

    sq.execute()  # type: ignore

    return sq.dataframe


def execute_bql_batch(queries: Iterable[str], subtype: str = "bql") -> Iterable[DataFrame]:
    """Runs the BQL queries as a batch first"""
    subqueries = []
    for query in queries:
        iqc = IqlQueryContainer(query, db=None)  # type: ignore
        subqueries_q = [q.bqlquery for q in iqc.get_subqueries_by_extension(subtype, None)]  # type: ignore
        subqueries += subqueries_q

    logger.debug("subqueries=%s", subqueries)

    # TODO: Fix this to work in execute_bql_batch
    # execute_bql_batch(subqueries)
    execute_bql_str_list_async_q(subqueries)

    results = [execute(q) for q in queries]
    return results  # type: ignore


def register(keyword: str):
    global _KEYWORD
    _KEYWORD = keyword
    extension = BqlExtension(keyword=keyword)
    register_extension(extension)

    extension2 = BqlExtension(keyword=keyword, subword="large")
    # Large mode: Fewer concurrent connections, writing to disk rather than keeping in memory
    extension2.use_path_replacement = lambda: True
    register_extension(extension2)

    extension3 = BqlExtension(keyword=keyword, subword="cache")
    extension3.use_path_replacement = lambda: False
    extension3.cache = cache.MemoryAndFileCache(max_age=3600 * 18, return_pyarrow_table=False)
    register_extension(extension3)

    extension4 = BqlExtension(keyword=keyword, subword="mcache")
    extension4.cache = cache.MemoryCache(max_age=None)
    register_extension(extension4)



# Copyright (C) 2025, IQMO Corporation [support@iqmo.com]
# All Rights Reserved

import logging
import os
from dataclasses import dataclass

import pymssql
from pandas import DataFrame

from ... import IqlExtension, SubQuery, register_extension
from ...datamodel import cache
from ...ql import substitute_replace

logger = logging.getLogger(__name__)

# https://github.com/pymssql/pymssql
# https://pymssql.readthedocs.io/en/stable/


# The connection cache is either connections **or** options to create a connection
_CONNECTION_CACHE: dict[str, pymssql.Connection | dict] = {}
CACHE_INVALIDATION_QUERY = None

def _get_connection_from_options(**kwargs) -> pymssql.Connection:
    return pymssql.connect(**kwargs)


def _get_connection(connection_name: str) -> pymssql.Connection:
    existing_conn: pymssql.Connection | dict = _CONNECTION_CACHE[connection_name]
    if isinstance(existing_conn, dict):
        return _get_connection_from_options(**existing_conn)
    else:
        return existing_conn


def _execute_query(conn: pymssql.Connection, query: str, parameters: dict | None = None) -> DataFrame:
    logger.debug("Executing query: %s", query)

    with conn.cursor(as_dict=True) as cursor:
        # dont perform substitutions, we do this with parameters

        query_sub = substitute_replace(query, {})
        logger.info("Running query %s", query_sub)
        cursor.execute(query_sub, parameters)
        results = cursor.fetchall()
        columns = cursor.description

        while cursor.nextset():
            results = cursor.fetchall()
            columns = cursor.description

    if results is not None and len(results) == 0:  # type: ignore
        return DataFrame(results, columns=[col[0] for col in columns])  # type: ignore
    return DataFrame(results)


@dataclass
class SqlServerExtensionPyMssqlConnect(IqlExtension):
    def executeimpl(self, sq: SubQuery) -> DataFrame:
        connection_name: str = sq.options.get("name", "default")  # type: ignore

        eager = sq.options.get("eager", False)
        conn_options = {k: v for k, v in sq.options.items() if k != "name" and k != "eager"}

        if eager:
            _CONNECTION_CACHE[connection_name] = _get_connection_from_options(**conn_options)
        else:
            _CONNECTION_CACHE[connection_name] = conn_options

        return DataFrame({"success": [True], "message": ["Connection Successful"]})


@dataclass
class SqlServerExtensionPyMssql(IqlExtension):
    @cache.iql_cache
    def executeimpl(self, sq: SubQuery) -> DataFrame:
        connection_name: str = sq.options.get("name", "default")  # type: ignore

        conn: pymssql.Connection = _get_connection(connection_name)

        query: str = sq.get_query()  # type: ignore

        parameters: dict = sq.options.get("PARAMETERS", None)  # type: ignore

        return _execute_query(conn, query, parameters=parameters)


def register(keyword: str):
    if CACHE_INVALIDATION_QUERY is not None:
        logger.info("Using CACHE_INVALIDATION_QUERY")
        ext_cache = cache.QueryInvalidationCache(CACHE_INVALIDATION_QUERY)
    else:
        ext_cache = None

    extension = SqlServerExtensionPyMssqlConnect(keyword=keyword, subword="pymssql_connect", cache=None)
    register_extension(extension)

    extension = SqlServerExtensionPyMssql(keyword=keyword, subword="pymssql", cache=ext_cache)
    register_extension(extension)

    extension = SqlServerExtensionPyMssql(keyword=keyword, cache=ext_cache)
    register_extension(extension)

    extension = SqlServerExtensionPyMssql(keyword=keyword, subword="nocache", cache=None)
    register_extension(extension)


import logging
from dataclasses import dataclass

import pandas as pd

from ..datamodel.extension import IqlExtension
from ..datamodel.subquery import SubQuery
from ..ql import _EXTENSIONS, register_extension

logger = logging.getLogger(__name__)


@dataclass
class CacheClearAll(IqlExtension):
    """Clears the entire cache"""

    keyword: str

    def executeimpl(self, sq: SubQuery) -> pd.DataFrame:
        for e in _EXTENSIONS.values():
            e.cache.clear_all()

        return pd.DataFrame([{"result:": True}])


@dataclass
class CacheClear(IqlExtension):
    """Clears a single item from the cache"""

    keyword: str

    def executeimpl(self, sq: SubQuery) -> pd.DataFrame:
        key: str = sq.options["key"]  # type: ignore
        for e in _EXTENSIONS.values():
            e.cache.clear(key)

        return pd.DataFrame([{"result:": True}])


def register(keyword):
    register_extension(CacheClearAll(keyword=keyword, subword="clear_all"))
    register_extension(CacheClear(keyword=keyword, subword="clear"))



def load_ipython_extension(ip, enable_autocomplete: bool = True):
    """Defer imports for non-ipython environments"""
    from IPython.core.getipython import get_ipython

    from .autocompletion_v2 import init_completer
    from .iql_magic import IqlMagic

    ip = get_ipython()

    if enable_autocomplete:
        init_completer(ip)

    ip.register_magics(IqlMagic)  # type: ignore
